#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ..ressource_classes import EditorAdaptor, adaptionmethod, \
    SingleResource, resource_property, resource_func


class mp3tagAdaptor(EditorAdaptor):
    save_in_place = True
    
    
class mp3tagBase(SingleResource, mp3tagAdaptor.AdaptiveClass):
    allowed_file_extensions = [".mp3"]
    
    
    title = resource_property("title")
    genre = resource_property("genre")
    artist = resource_property("artist")
    album = resource_property("album")
    
    
    
    def remove_tag_from_genre(self, tag):
        ng = []
        for i in self.genre.split(" "):
            if tag not in i: ng.append(i)
        self.genre = " ".join(ng).strip()
        self.save()
            
    
    


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .ressource_classes import *
from .imagepower import *
from .mp3power import *
from datetime import datetime
from contextlib import redirect_stdout


class Image(ImageBase):
    adaptor = ImageAdaptor(_primary_name="Image.adaptor")
    
    
class mp3tag(mp3tagBase):
    adaptor = mp3tagAdaptor(_primary_name="mp3tag.adaptor")


from .FilelistCreator import FilelistCreator


class PlaylistCreator(FilelistCreator):
    editor_class = mp3tag
    filelist_extension = ".m3u"
    mytag = ""
    #file_start = "#EXTM3U"
    
    @classmethod
    def adapt(cls, *args, **kwargs):
        return cls.editor_class.adapt(*args,**kwargs)
    
    def print_mp3_tags(self, name, header=True):
        #os.makedirs(dest_folder)
        if isinstance(name, self.Filelist): name = name.name
        l = self.imported_lists[name]
        self._create_print(l, name, header=header)
        
    def print_tags_to_file(self, l, destpath, playlist_name="", header=True,
            txt_name=None):
        if txt_name is None: txt_name = playlist_name
        with open(os.path.join(destpath, txt_name + ".txt"), "w") as f:
            with redirect_stdout(f):
                self._create_print(l, playlist_name, header)
        
    def _create_print(self, l, name="", header=True):
        if header:
            print("-------------------------------------------------")
            print(f"Playlist '{name}' containing {len(l)} entries")
            print("-------------------------------------------------")
            print("Date: ", datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
            print()
            print("Title / Artist / Album / 'filename'")
            print()
        for file in l:
            with self.editor_class(file) as obj:
                print(obj.title," / ",obj.artist, " / ",obj.album, " / '",
                        os.path.basename(file),"'", sep="")
    
    def copy_files(self, filelist, dest_folder, numerate=False,
            clear_tags=True, include_txt=False):
        if isinstance(filelist, self.Filelist): filelist = filelist.name
        created_files = super().copy_files(filelist, dest_folder, numerate)
        if clear_tags and self.mytag:
            print("Removing custom tags.")
            for file in os.listdir(dest_folder):
                if not self._file_allowed(file): continue
                with self.editor_class(path.join(dest_folder, file)) as obj:
                    obj.remove_tag_from_genre(self.mytag)
        if include_txt:
            self.print_tags_to_file(created_files,dest_folder,
                    filelist, txt_name="0_readme")
    


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.launcher import launch
from .. import DependencyManager


with DependencyManager(__name__) as tester:
    tester.test_shellcommand("yad", pkg="yad")




def start(IconObject):
    cmd = tuple(create_yad_cmd(IconObject))
    #print(cmd)
    child = launch(cmd, stdout=True)
    launch.thread(waiter, IconObject, child.stdout)
    return child.pid

def create_yad_cmd(IconObject):
    # the default behavior when clicking on the icon is to close the  Can be
    # changed with click_cmd
    # if menu=="":
    #            menu=standard_menu
    yield "yad"
    yield "--notification"
    yield "-- listen"
    if IconObject.path:
        yield "--image=" + IconObject.path
    #if click_cmd:
     #   yield " --command=" + shlex.quote(click_cmd)
    if IconObject.tooltip:
        yield "--text=" + IconObject.tooltip
    items = IconObject.items
    if items:
        text="--menu="
        for item in items:
            text+= item[0] +"!echo "+item[0]+"|"
        text=text[:-1]
        yield text


def waiter(Iconobject, stdout):
    for line in stdout:
        # this is executed, each time we receive a message from YAD
        for i in Iconobject.items:
            if line == i[0] + "\n":
                # print("execute: "+str(i[1]))
                launch.thread(i[1])
    if Iconobject.left_click:
        Iconobject.left_click()


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import DependencyManager

with DependencyManager(__name__) as manager:
    pynput = manager.import_module("pynput")

mouse = pynput.mouse.Controller()

def pos():
    return mouse.position

move=mouse.move


def moveto(x,y):
    mouse.position = (x,y)


names = {i : pynput.mouse.Button(i) for i in range(1,31)}
#this is necessary because somehow pynput send button.value instead of button

#click=mouse.click
press=mouse.press
rls=mouse.release

def scroll(vertical,horizontal):
    return mouse.scroll(horizontal,vertical)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from warnings import warn
from . import NamedKey, keyb, NamedButton
from ..winapps import pythoneditor
from .event_classes import StringAnalyzer, EventCombination, EventSequence, \
    StringEvent
from .hookpower import HookAdaptor, CallbackHook, KeyhookBase, ButtonhookBase
from Dhelpers.launcher import launch
from Dhelpers.baseclasses import TimedObject, KeepInstanceRefs
from Dhelpers.arghandling import check_type
from Dhelpers.counting import dpress
import collections, inspect
from os import path

class PatternListener:
    
    max_thread_num = 1
    
    def __init__(self, buffer = 4):
        self.eventdict = dict()
        self.recent_events = collections.deque()
        self.buffer = buffer
        self.blocked_hks = []
        self.stringevent_analyzer = StringAnalyzer(NamedKey, NamedButton)
        self.reset_analyzefunc()
        self.active_thread_num = 0
        
    
    def event(self, k):
        if not self.eventdict: return
        # _print(k)
        if self.active_thread_num >= self.max_thread_num: return
        self.recent_events.append(k)
        if len(self.recent_events) > self.buffer: self.recent_events.popleft()
        recent_events = self.recent_events.copy()
        for event, action in self.eventdict.items():
            members = event.members
            lm = len(members)
            if lm > len(recent_events): continue
            # if more members are required than actual events passed
            for i in range(-1, -lm - 1, -1):
                if members[i] != recent_events[i]: break
            else:
                launch.thread(self.runscript, action, event)
                # this means that for each member event, the suiting recent
                # event has been found
                self.recent_events.clear()

    def runscript(self, action, hk):
        if self.active_blocks: return
        if self.analyze_func:
            if self.analyze_func(action, hk) is True: self.reset_analyzefunc()
            return
        assert self.active_thread_num >= 0
        if self.active_thread_num >= self.max_thread_num: return
        self.active_thread_num += 1
        try:
            # print(hk_func)
            # print(hk,hk_func,type(hk_func))
            if type(action) is str:
                if action.startswith("[dkeys]"):
                    if dpress(hk, 0.15):
                        keyb.send(action[7:], delay=1)
                else:
                    keyb.send(action)
            elif callable(action):
                # the following makes sure that the hk_func is accepting 1
                # argument even if the underlying func does not
                try:
                    return action(hk)
                except TypeError:
                    pass
                return action()
            else:
                raise TypeError
        finally:
            self.active_thread_num -= 1
        
        
    def set_analyzefunc(self, func, timeout=5):
        assert callable(func)
        self.analyze_func = func
        launch.thread(self.reset_analyzefunc, initial_time_delay=timeout)
        
    def reset_analyzefunc(self):
        self.analyze_func = None
        
    def show_source_code(self, timeout=5):
        self.set_analyzefunc(self._jump_to_action, timeout=timeout)
    
    @staticmethod
    def _jump_to_action(action, hk):
        file = path.abspath(inspect.getsourcefile(action))
        linenumber = inspect.getsourcelines(action)[-1]
        pythoneditor.jump_to_line(file, linenumber)
        return True
    

    def add_event(self, event, action):
        if event in self.eventdict:
            raise ValueError(f"Tiggerevent {event} defined more than one time.")
        self.eventdict[event] = action

    def add_sequence(self, string, action):
        event = self.stringevent_analyzer.create_from_str(
                string).hotkey_version()
        self.add_event(event, action)

    def add_hotkey(self, string, action, rls=False, block=True):
        event = self.stringevent_analyzer.create_from_str(string)
        if isinstance(event, EventSequence): raise ValueError
        if isinstance(event, StringEvent):
            if not event.press: raise ValueError
            if block: self.blocked_hks.append(event)
            if rls: event += event.reverse()
        elif isinstance(event, EventCombination):
            event = event.hotkey_version(rls=rls)
        else:
            raise TypeError
        self.add_event(event, action)

    # A decorator
    def sequence(self, *strings):
        def decorator(decorated_func):
            for string in strings: self.add_sequence(string, decorated_func)
            return decorated_func
        return decorator

    # A decorator
    def hotkey(self, *strings, rls=False, block=True):
        def decorator(decorated_func):
            for string in strings:
                self.add_hotkey(string, decorated_func, rls=rls, block=block)
            return decorated_func
        return decorator

    def hotkey_rls(self, *strings, block=True):
        return self.hotkey(*strings, rls=True, block=block)

    def add_triggerdict(self, triggerdict):
        for eventstring, action in triggerdict.items():
            self.add_sequence(eventstring, action)

    

    active_blocks = 0

    @classmethod
    def block(cls):
        cls.active_blocks += 1  # print(cls.active_blocks)

    @classmethod
    def unblock(cls, delay=None):
        if delay:
            launch.thread(cls._unblock, initial_time_delay=delay)
        else:
            cls._unblock()

    @classmethod
    def _unblock(cls):
        cls.active_blocks -= 1
        # print(cls.active_blocks)
        if cls.active_blocks < 0: raise ValueError

    @classmethod
    def paused(cls, timeout=10):
        return PauseObject(timeout, cls)
        # this creates a PauseObject suitable for a with statement
        # usage: with instance.paused():   or    with TriggerManager.paused():



class RegisteredHook(PatternListener):
    
    def __init__(self, buffer, hook_instance, container_triggerman=None):
        super().__init__(buffer)
        check_type(CallbackHook, hook_instance)
        if isinstance(hook_instance, KeyhookBase):
            self.stringevent_analyzer = NamedKey.Event
        elif isinstance(hook_instance, ButtonhookBase):
            self.stringevent_analyzer = NamedButton.Event
        self.hook_instance = hook_instance(self.event)
        self.triggerman_instance = container_triggerman
        
    def event(self, k):
        super().event(k)
        if self.triggerman_instance: self.triggerman_instance.event(k)
        
    def start(self):
        if self.blocked_hks or self.triggerman_instance and \
                self.triggerman_instance.blocked_hks:
            try:
                self.hook_instance = self.hook_instance(reinject_func =
                                self.reinject_func )
            except NotImplementedError:
                if self.blocked_hks: raise
        return self.hook_instance.start()
    
    def stop(self):
        return self.hook_instance.stop()

    def reinject_func(self, event_obj):
        if event_obj in self.blocked_hks: return False
        if self.triggerman_instance and event_obj in \
                self.triggerman_instance.blocked_hks: return False
        return True
    
    @property
    def analyze_func(self):
        if self._analyze_func: return self._analyze_func
        if self.triggerman_instance: return self.triggerman_instance.analyze_func
        
    @analyze_func.setter
    def analyze_func(self, val):
        self._analyze_func = val
        
    def reset_analyzefunc(self):
        try:
            if self._analyze_func is None:
                if self.triggerman_instance and self.triggerman_instance.analyze_func:
                    self.triggerman_instance.analyze_func = None
        except AttributeError:
            pass
        self.analyze_func = None

class TriggerManager(PatternListener,TimedObject, HookAdaptor.AdaptiveClass,
        KeepInstanceRefs):
    
    adaptor = HookAdaptor(group="triggerman",
            _primary_name="TriggerManager.adaptor")
    
    
    def __init__(self, timeout=60, buffer=4):
        PatternListener.__init__(self,buffer=buffer)
        TimedObject.__init__(self,timeout=timeout)
        KeepInstanceRefs.__init__(self)
        self.registered_hooks = []
        self.was_started = False
        self.max_thread_num = TriggerManager.max_thread_num


    @classmethod
    def start_all(cls):
        for inst in cls.get_instances():
            if inst.was_started is False: inst.start()
            
    def add_hook(self, hook_instance, buffer=None,**hook_kwargs):
        timeout = None if self.timeout is None else self.timeout + 5
        hook_instance = hook_instance(timeout=timeout, **hook_kwargs)
        buffer = self.buffer if buffer is None else buffer
        new_registered_hook = RegisteredHook(buffer, hook_instance,
                container_triggerman=self)
        self.registered_hooks.append(new_registered_hook)
        return new_registered_hook
    
    def hook_keys(self, backend=None, **hook_kwargs):
        if backend:
            adaptor = self.adaptor_class(keys=backend)
        else:
            adaptor = self.adaptor
        return self.add_hook(adaptor.keys(),**hook_kwargs)
        
    def hook_buttons(self,backend=None,**hook_kwargs):
        if backend:
            adaptor = self.adaptor_class(buttons=backend)
        else:
            adaptor = self.adaptor
        return self.add_hook(adaptor.buttons(), **hook_kwargs)
    
    def hook_custom(self, backend=None,**hook_kwargs):
        if backend:
            adaptor = self.adaptor_class(custom=backend)
        else:
            adaptor = self.adaptor
        return self.add_hook(adaptor.custom(), **hook_kwargs)
    
    def _start_action(self):
        self.was_started = True
        for rhook in self.registered_hooks: rhook.start()
        
    def _stop_action(self):
        for rhook in self.registered_hooks:
            try:
                rhook.stop()
            except Exception as e:
                warn(e)


    def _timeout_action(self):
        warn("Timeout after %s seconds: Stopping TriggerManager %s." % (
            self.timeout,self))



    @property
    def max_thread_num(self):
        return self._max_thread_num
    
    @max_thread_num.setter
    def max_thread_num(self, val):
        self._max_thread_num = val
        for rhook in self.registered_hooks:
            rhook.max_thread_num = val


class PauseObject(TimedObject):
    def __init__(self, timeout, cls):
        super().__init__(timeout=timeout)
        self.cls = cls
    
    def _start_action(self):
        self.cls.block()
        
    def _stop_action(self):
        self.cls.unblock(delay=0.1)
        # delaying the unblock is necessary because when the user types,
        # the key up events are often catched without intention otherwise


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import os, warnings, sys, random, functools, shutil
from collections import defaultdict
from contextlib import contextmanager, redirect_stdout
path=os.path

from Dhelpers.boolean_function import BooleanFunction


class Filelist(BooleanFunction):
    creator = None
    def __init__(self, name=None, func=None, add=True, **kwargs):
        self.last_found_paths=[]
        if name is None: add = False
        if not func:
            if not kwargs: raise ValueError
            func = lambda obj: self.creator.default_selection_func(obj, **kwargs)
        BooleanFunction.__init__(self, name, func)
        if add: self.creator.add_inst(self)


class FilelistCreator:
    editor_class = None
    filelist_extension = ""
    file_start = ""
    allowed_file_extensions = "inherit"
    Filelist = None
    

    def _file_allowed(self, file):
        if self.allowed_file_extensions == "inherit":
            allowed_ext =  self.editor_class.allowed_file_extensions
        else:
            allowed_ext = self.allowed_file_extensions
        ext = os.path.splitext(file)[-1]
        if ext in allowed_ext: return True
        return False
    
    def __init__(self, basepath=None, destpath=None,insert_initial=None):
        filelist_cls = type(f"Filelist", (Filelist,), {})
        filelist_cls.creator = self
        self.Filelist = filelist_cls
        self.basepath = basepath
        self.destpath = destpath
        self.filelist_objs = {}
        self.file_paths = {}
        self.imported_lists = None
        self.filelist_combinations = list()
        self._custom_file_func = NotImplemented
        self.insert_initial = insert_initial
        
    @property
    def destpath(self):
        if not self._destpath and self.basepath:
            parent = os.path.split(self.basepath)[0]
            return os.path.join(parent, "__Playlists")
        return self._destpath
        
    @destpath.setter
    def destpath(self, val):
        self._destpath = val
        
    
    def add(self, name, func=None, **kwargs):
        new_inst = self.Filelist(name,func=func, **kwargs)
        self.add_inst(new_inst)
    
    def add_inst(self, filelist_inst):
        assert isinstance(filelist_inst, self.Filelist)
        name = filelist_inst.name
        self.filelist_objs[name] = filelist_inst
        self.file_paths[name] = filelist_inst.last_found_paths  #shortcut

    @staticmethod
    def default_selection_func(obj, case_sensitive=False, exact_match=False,
            **kwargs):
        for attr_name, value in kwargs.items():
            string = str(getattr(obj, attr_name))
            value = str(value)
            if not case_sensitive:
                value = value.lower()
                string = string.lower()
            if exact_match:
                if value != string: return False
            else:
                if value not in string: return False
        return True
    
    def find_files(self, *names, suppress_error=True, warn=True):
        for dirpath, _, files in os.walk(self.basepath):
            with self.Filelist.use_cache() as reset_cache:
                for file in files:
                    if not self._file_allowed(file): continue
                    file = os.path.realpath(os.path.join(dirpath, file))
                    try:
                        with self.editor_class(file) as obj:
                            for name in names:
                                filelist_obj = self.filelist_objs[name]
                                if filelist_obj(obj): yield name, file
                    except Exception as e:
                        if not suppress_error: raise
                        if warn:
                            warnings.warn(f"\nerror with file {file}:\n{e}")
                            print()
                    finally:
                        reset_cache()
    
    def assemble_lists(self, *names, **kwargs):
        if not names: names = self.file_paths.keys()
        for name in names:
            self.file_paths[name].clear()
        for name, file in self.find_files(*names, **kwargs):
            # print(name,file)
            self.file_paths[name].append(file)
    
    
    def write_paths(self, *names, assemble=True, insert_initial=None,
            overwrite = True, **kwargs):
        if assemble: self.assemble_lists(*names, **kwargs)
        print("Creating playlists:")
        os.makedirs(self.destpath, exist_ok=True)
        for name, file_path_list in self.file_paths.items():
            self._write_filelist(name,file_path_list, self.destpath,
                    insert_initial=insert_initial, overwrite=overwrite)
            
            
    def _write_filelist(self, name, file_path_list, path, overwrite=True,
            insert_initial = None):
        if insert_initial is None: insert_initial = self.insert_initial
        fpath_base = os.path.join(path, name)
        fpath = fpath_base + self.filelist_extension
        if overwrite is False:
            num = 0
            while os.path.isfile(fpath):
                num += 1
                fpath = fpath_base + "__" + str(num) + self.filelist_extension
        with open(fpath, "w") as new:
            new.write(self.file_start + "\n")
            if insert_initial:
                new.write(self.get_custom_file(insert_initial)+"\n")
            for file in file_path_list:
                new.write(file + "\n")
        print(f"{name}: {len(file_path_list)} songs in {fpath}")



    def import_lists(self, destpath=None, add_ext=()):
        if destpath is None: destpath = self.destpath
        allowed_ext = (self.filelist_extension,) + add_ext
        self.imported_lists = defaultdict(list)
        for file in os.listdir(destpath):
            name, ext = os.path.splitext(file)
            if ext not in allowed_ext: continue
            with open(os.path.join(destpath,file), "r") as opened:
                last_line = ""
                while last_line == "": last_line = opened.readline().strip()
                if self.insert_initial:
                    initial_lines = self.get_custom_file(
                            self.insert_initial).split("\n")
                    for in_line in initial_lines:
                        if in_line.strip() != last_line:
                            raise ValueError(in_line, last_line)
                        last_line = opened.readline().strip()
                        #this will pop out the initial lines before importing
                self.imported_lists[name].append(last_line)
                for line in opened.readlines():
                    self.imported_lists[name].append(line.strip())
        return self.imported_lists
    
    def define_combination(self, *args, **kwargs):
        c = FilelistCombination(self, *args,**kwargs)
        self.filelist_combinations.append(c)
        return c
    
    def write_combinations(self, *args,**kwargs):
        for c in self.filelist_combinations: c.write(*args,**kwargs)
        
    def get_custom_file(self, name_or_number):
        if self._custom_file_func is NotImplemented: raise NotImplementedError
        file = self._custom_file_func(self, name_or_number)
        if not os.path.isfile(file): raise FileNotFoundError(file)
        return file
    
    def custom_file_finder(self, func):
        """A decorator to define the function returning files for
        get_custom_file()"""
        self._custom_file_func = func
        return func

        
    def copy_files(self, filelist, dest_folder, numerate=False):
        os.makedirs(dest_folder)
        if isinstance(filelist, self.Filelist): filelist=filelist.name
        if isinstance(filelist, str):
            filelist = self.imported_lists[filelist]
        print(f"Copying {len(filelist)} files to {os.path.realpath(dest_folder)} ...")
        num = 0
        new_names = []
        for file in filelist:
            new_name = path.split(file)[1]
            if numerate:
                num += 1
                new_name = f"{num}__{new_name}"
            new_full_name = path.join(dest_folder, new_name)
            shutil.copy(file, new_full_name)
            new_names.append(new_full_name)
        print("Done.")
        return new_names

class FilelistCombination:
    
    def __init__(self, playlist_creator, name, file_list_names,*, insert=None,
            insert_initial=None, from_imported=True, destpath=None,
            avoid_double=True):
        # insert can be a file path or name of a previously defined
        # FileList. It will be inserted in between every entry. (Useful for
        # silence between music tracks e.g.)
        self.name = name
        self.creator = playlist_creator
        assert isinstance(file_list_names, (tuple, list))
        self.file_lists = file_list_names
        self.insert = insert
        self.insert_initial = insert_initial
        self.from_imported = from_imported
        self.destpath = destpath
        self.avoid_double = avoid_double
        
    def _process_fileinfo(self, list_info, defined_lists):
        if isinstance(list_info, self.creator.Filelist):
            list_info = list_info.name
            if not list_info: raise ValueError
        if list_info in defined_lists:
            return random.choice(defined_lists[list_info])
            #first, check if a FileList has been defined with that name
            # if not, then try to interpret the given name as an argument for
            # custom_file_finder
        else:
            return self.creator.get_custom_file(list_info)
    
    def create(self, insert=None, insert_initial=None, from_imported=None):
        if insert is None: insert = self.insert
        if insert_initial is None: insert_initial = self.insert_initial
        if from_imported is None: from_imported = self.from_imported
        creator = self.creator
        lists = creator.imported_lists if from_imported else creator.file_paths
        seq = list()
        clean_seq = list()
        if insert_initial:
            seq.append(self._process_fileinfo(insert_initial, lists))
        for name in self.file_lists:
            file = self._process_fileinfo(name, lists)
            if file:
                while self.avoid_double and file in clean_seq:
                    # by default, avoid picking the same file twice
                    file = self._process_fileinfo(name, lists)
                seq.append(file)
                clean_seq.append(file)
            if insert:
                if os.path.isfile(insert):
                    seq.append(insert) #just return the input, in case it is a
                    # explicit path
                else:
                    seq.append(self._process_fileinfo(insert, lists))
        return seq, clean_seq
    
    def write(self, name=None, destpath=None, overwrite=True,
            copy=False,**kwargs):
        name = name if name else self.name
        if not destpath:
            destpath = self.destpath if self.destpath else "combinations"
        if not path.isabs(destpath):
            destpath = path.join(self.creator.destpath, destpath)
        destpath = path.join(destpath, name)
        copy_folder = name+"_files"
        filelist, clean_filelist = self.create(**kwargs)
        #print(filelist)
        os.makedirs(destpath, exist_ok=True)
        writelist = filelist
        if copy is True:
            writelist = [path.join(copy_folder,path.split(file)[1]) for file in
                filelist]
        self.creator._write_filelist(name, writelist, destpath,
                overwrite=overwrite, insert_initial = False)
        
        try:
            print_command = self.creator.print_tags_to_file
        except AttributeError:
            pass
        else:
            print_command(clean_filelist, destpath, name, header=True)
                    
        if copy is True:
            copypath = path.join(destpath, copy_folder)
            if os.path.isdir(copypath): shutil.rmtree(copypath)
            self.creator.copy_files(filelist, copypath, numerate=False)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from Dhelpers.baseclasses import AdditionContainer
from Dhelpers.arghandling import check_type
from ..event_classes import StringEvent, EventSequence, EventCombination
from .. import Adaptor, adaptionmethod, AdaptionError, Layout
import functools, time, warnings

sleep = time.sleep



def _doc(middle):
    return "*Class attribute.* Default time (in milliseconds) to wait between" \
           f" sending {middle}. You can set a custom value to this attribute"\
            " for the whole class or for a single instance."

def default_delay_doc(obj_name):
    print(_doc(f"several {obj_name} in a row"))

def default_duration_doc(obj_name):
    print(_doc(f"press and release event of the same {obj_name}"))



class EventObjectSenderMixin:
    default_delay = None
    
    def _send_event(self, event, **kwargs):
        raise NotImplementedError
    
    def send_event(self, *events, delay=None, **kwargs):
        delay = self.default_delay if delay is None else delay
        for event in events:
            if isinstance(event, EventSequence):
                self.send_event(*event.members, **kwargs)
            elif isinstance(event, EventCombination):
                self.send_event(event.sending_version(), **kwargs)
            else:
                try:
                    name = event.name
                except AttributeError:
                    pass
                else:
                    if name.startswith("[") and name.endswith("]"):
                        name = name[1:-1]
                        event.name = name
                try:
                    self._send_event(event, **kwargs)
                except TypeError:
                    raise TypeError(f"event argument {event} not allowed for "
                                    f"send_event method of object {self}.")
                if delay: sleep(delay/1000)
                
                
                

class Sender(AdditionContainer.Addend):
    
    #only_defined_names = False
    default_delay = None  # must be set in subclass
    custom_text_method = True
    
    def get_backend_name(self, name):
        return name
    
    def press(self, *names, delay=None):
        if delay is None: delay = self.default_delay
        for k in names:
            self._press(k)
            if delay: time.sleep(delay/1000)
        
        
    def _press(self, name, send_infos=(), expand_shifted = False):
        send_infos = list(send_infos) #creates an independent copy
        backend_name = self.get_backend_name(name)
        if isinstance(backend_name, Layout.ShiftedKey):
            if expand_shifted:
                for bn in self._expand_shifted(backend_name):
                    self._press_action(bn)
                    send_infos.append((bn, self))
                return send_infos
            else:
                backend_name = [backend_name.keyname]
        self._press_action(backend_name)
        send_infos.append((backend_name, self))
        return send_infos

    def _press_action(self, name):
        raise NotImplementedError

    def text(self, text, delay=None, custom=None, **kwargs):
        if delay is None: delay = self.default_delay
        if custom is None: custom = self.custom_text_method
        for character in text:
            if custom:
                try:
                    self._text_action(character, **kwargs)
                    if delay: time.sleep(delay/1000)
                except NotImplementedError:
                    custom = False
            if not custom:
                self.tap(character, delay=delay, duration=10)
                

    def _text_action(self, character, **kwargs):
        raise NotImplementedError
    
    def tap(self, character, delay=None, duration=None):
        # defined for compatability
        # duration attribute is ignored because no rls defined
        self.press(character, delay=delay)
    
    def send_eventstring(self, string, auto_rls=True, delay=None):
        #auto_rls attribute is ignored because no rls defined
        for entry in StringEvent.split_str(string):
            if isinstance(entry, str):
                self.press(entry, delay=delay)
            else:
                raise TypeError
        
    _starting_symbol = "<"
    _ending_symbol = ">"
    
    
    def send(self, s, auto_rls=True, delay=None, **text_kwargs):
        text_kwargs["delay"] = delay
        _combination = self._ending_symbol + self._starting_symbol
        while True:
            s0, s1, s2 = s.partition(self._starting_symbol)
            if s0 != "": self.text(s0, **text_kwargs)
            if s2 == "": break
            if s2.startswith(_combination):
                # if <>< is typed, send a single < and search if any more <
                # are present
                self.text(self._starting_symbol,**text_kwargs)
                s = s2[2:]
            else:
                t0, t1, t2 = s2.partition(self._ending_symbol)
                if t1 == "":
                    # this happens if ">" was not found so that < is not closed.
                    self.text(self._starting_symbol + t0, **text_kwargs)
                    break
                elif t0 != "":
                    self.send_eventstring(t0, auto_rls=auto_rls, delay=delay)
                if t2 == "": break
                s = t2



class PressReleaseSender(Sender):
    
    default_duration = None

    def rls(self, *names, delay=None):
        if delay is None: delay = self.default_delay
        for k in reversed(names):
            self._rls(k)
            if delay: time.sleep(delay/1000)
            
    def _rls(self, name):
        backend_name = self.get_backend_name(name)
        if isinstance(backend_name, Layout.ShiftedKey):
                backend_name = backend_name.keyname
        self._rls_action(backend_name)
    
    def _rls_action(self, name):
        raise NotImplementedError
    
    def send_eventstring(self, string, auto_rls=True, delay=None):
        for entry in StringEvent.split_str(string):
            if isinstance(entry, str):
                if entry.endswith("_rls"):
                    if auto_rls: raise ValueError
                    self.rls(entry[:-4],delay=delay)
                elif auto_rls:
                    self.tap(entry, delay=delay)
                else:
                    self.press(entry,delay=delay)
            elif isinstance(entry, (list,tuple)):
                self.comb(*entry, delay=delay)
            else:
                raise TypeError(entry)
            
            
    def comb(self, *names, delay=None, duration=None, expand_shifted=False):
        if delay is None: delay = self.default_delay
        if duration is None: duration = self.default_duration
        delay = delay/1000
        duration = duration / 1000
        send_infos = []
        try:
            for name in names:
                send_infos = self._press(name, send_infos, expand_shifted)
                sleep(delay)
            sleep(duration)
        finally:
            e = None
            for name, sender_inst in reversed(send_infos):
                try:
                    sender_inst._rls_action(name)
                except Exception as e:
                    continue
                else:
                    sleep(delay)
            if e: raise e

    def tap(self, *names, delay=None, duration=None, repeat=1):
        for _ in range(repeat):
            for n in names:
                PressReleaseSender.comb(self,n, delay=delay,
                        duration=duration, expand_shifted=True)

    @functools.wraps(Sender.press)
    def pressed(self, *names, **kwargs):
        return PressedContext(self, *names, **kwargs)


class PressedContext:
    
    def __init__(self, sender_instance, *names, delay=None):
        self.sender_instance = sender_instance
        self.names = names
        if not delay: delay = sender_instance.default_delay
        self.delay = delay/1000
        self.send_infos = []
    
    def __enter__(self):
        for name in self.names:
            self.send_infos = self.sender_instance._press(name, self.send_infos)
            sleep(self.delay)
        return self.sender_instance
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        #if exc_type == AdaptionError: return  # just raise this error directly
        e = None
        for name, sender_inst in reversed(self.send_infos):
            try:
                sender_inst._rls_action(name)
            except Exception as e:
                continue
            else:
                sleep(self.delay)
        if e:
            if exc_type == type(e): return  # only reraise press error
            raise



class AdaptiveSender(Sender, Adaptor):
    baseclass = True
    NamedClass = None
    #stand_dicts = defaultdict(lambda : defaultdict(lambda : list()))
    translation_dict = None
    default_translations = {}  #contains mapping of target space to instance
                    # defining the default translations
    _effective_dict = {}
    
    @property
    def Event(self):
        return self.NamedClass.Event
    
    
    def get_backend_name(self, name):
        try:
            return self._effective_dict.apply(name)
        except AttributeError:
            return self._effective_dict.get(name, name)
        

    @adaptionmethod("press", require=True)
    def _press_action(self, name):
        try:
            return self._press_action.target(name)
        except Exception as e:
            if isinstance(e,self._press_errortype): raise NameError(name) from e
            raise
    
    @_press_action.target_modifier
    def _press_tm(self, target):
        self._press_errortype = self._get_errortype(target)
        self.create_effective_dict(transfer=True)
        return target


    names = _press_action.adaptive_property("names", default={},
            ty=(dict, list, tuple))
    
    def _get_errortype(self,target):
        try:
            target("ASFASDFASDFASxcvjxcjsjsj23")
        except Exception as e:
            return type(e)
        raise RuntimeError("Name 'ASFASDFASDFASxcvjxcjsjsj23' is allowed!?")

    
    def set_translation_dict(self, dic, make_default=False):
        check_type(dict, dic)
        self.translation_dict = dic
        self.create_effective_dict(make_default=make_default)
    
    def _get_default_inst(self, target_space):
        enforced_inst = self.default_translations.get(target_space)
        if enforced_inst:
            assert isinstance(enforced_inst, self.__class__)
            assert enforced_inst._press_action.target_space is target_space
        return enforced_inst
    
    def create_effective_dict(self, make_default=False, transfer=False):
        ts = self._press_action.target_space
        if ts is None:
            if make_default: raise ValueError("Option make_default=True used, "
                  f"but no backend chosen yet for {self}." )
            return False
        enforced_inst = None
        if transfer: enforced_inst = self._get_default_inst(ts)
        self._create_effective_dict(ts, enforced_inst=enforced_inst)
        if make_default:
            # set this instance to be the default reference for this backend
            self.default_translations[ts] = self
            for instance in self.get_instances():
                if instance is not self and \
                        instance._press_action.target_space is ts:
                    instance.create_effective_dict(transfer=True)
    
    
    def _create_effective_dict(self, target_space, enforced_inst=None):
        #this should be executed:
        # - when amethod is adapted (i.e. target space changes)
        # - when NamedClass changes
        # - when name_translation changes
        # - when new enforced inst is set
        # stan dict format:
        #       NamedKey : Object to send to backend
        # translation_dict_format:
        #     user_input_name : remapped_name (may be ShiftedKey instance)
        #     (both sides of this are usually standardized automatically)
        names = self.names
        if self.NamedClass:
            stand_dict = self.NamedClass.StandardizingDict(names)
        else:
            stand_dict = names
        
        if enforced_inst: self.translation_dict = enforced_inst.translation_dict
        # simplify by using the same effective dict??
        
        if self.translation_dict:
            copy = stand_dict.copy()  # copy is standardizing dict too
            for key, remapped in self.translation_dict.items():
                stand_dict[key] = self._process_trans_item(remapped, copy)
                
        self._effective_dict = stand_dict
        
    
    def _process_trans_item(self, remapped, copy):
        try:
            return copy[remapped]  # check if standardized(remapped) is among
            # the backend names defined in self.names (copy is standardized
            # copy of it)
        except KeyError:
            return remapped #this is not standardized
            # this is just in case if the backend accepts keys not
            # defined in self.names
            
            
            
class AdaptivePressReleaseSender(PressReleaseSender, AdaptiveSender,
        EventObjectSenderMixin):
    baseclass = True



    @adaptionmethod("rls", require=True)
    def _rls_action(self, name):
        try:
            return self._rls_action.target(name)
        except Exception as e:
            if isinstance(e, self._rls_errortype):
                raise NameError(name) from e
            raise
    
    @_rls_action.target_modifier
    def _rls_tm(self, target):
        self._rls_errortype = self._get_errortype(target)
        self.create_effective_dict()
        return target


    def _send_event(self, event, reverse_press=False, autorelease=False):
        if not isinstance(event, self.Event): raise TypeError
        if event.press and not reverse_press or not event.press and reverse_press:
            self.press(event.name)
            if autorelease: self.rls(event.name)
        else:
            self.rls(event.name)
            
        
class AdaptivePRSenderShifted(AdaptivePressReleaseSender):
    baseclass = True
    
    use_shifted = AdaptiveSender._press_action.adaptive_property("use_shifted",
            False, ty=bool)
    
    def _process_trans_item(self, remapped, copy):
        sup = super()._process_trans_item
        if isinstance(remapped, Layout.ShiftedKey):
            # extract the keyname from the ShiftedKey object and compose it
            # again afterwards
            if not self.use_shifted:
                warnings.warn(f"Found ShiftedKey {remapped} although attribute"
                  f" use_shifted of {self} is set to {self.use_shifted}.")
            backend_obj = sup(remapped.keyname, copy)
            return Layout.ShiftedKey(backend_obj, num=remapped.num)
        else:
            return sup(remapped, copy)
    
    def _create_effective_dict(self, *args, **kwargs):
        super()._create_effective_dict(*args,**kwargs)
        self.shift_obj = self.get_backend_name("shift")
        self.atr_gr_obj = self.get_backend_name("altgr")
        
    def _expand_shifted(self, obj):
        if not isinstance(obj,Layout.ShiftedKey): return [obj]
        ret = []
        if obj.num in (1,3): ret.append(self.shift_obj)
        if obj.num >= 2: ret.append(self.atr_gr_obj)
        ret.append(obj.keyname)
        return ret
    
    def _effective_dict_expanded(self):
        return {i : self._expand_shifted(j) for i,j in \
                self._effective_dict.items(prefer_single=True)}



    
class CombinedSender(AdditionContainer, PressReleaseSender,
        EventObjectSenderMixin, basic_class=Sender):
    
    def __init__(self, *args):
        super().__init__(*args)
        self.default_delay = self._members[0].default_delay
        self.default_duration = self._members[0].default_duration
        
    @AdditionContainer.create_try_error_method(NameError)
    def _press(self, name):
        pass

    @AdditionContainer.create_try_error_method(NameError)
    def _rls(self, name):
        pass
    
    @AdditionContainer.create_try_error_method(TypeError)
    def _send_event(self, event, **kwargs):
        pass
    
    @AdditionContainer.create_try_error_method(NotImplementedError)
    def _text_action(self, character, **kwargs):
        pass
        
    def __call__(self, *args, **kwargs):
        return self.send(*args,**kwargs)

    # useful methods:
    # -- send_event: This will call EventObjectSender.send_event and then
    # iterate over all members' _send_event methods until no TypeError is raised
    # -- press and rls will iterate over _press_action and _rls_action
    # -- text will call super().text which will iterate over _text
    # -- send. This will automatically call prs, rls, text and thus iterate
    #       as well
    
    
    
    



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#


pynput = "pynput"
evdev = "evdev"
pystray = "pystray"

class linux:
    
    class HookAdaptor:
        default  = pynput
        triggerman = dict(buttons = pynput, keys = evdev, custom=evdev)
        keywait = triggerman
        
    class MouseAdaptor:
        default = pynput
        Dfuncs = default
        keywait = default
        
    class WindowAdaptor:
        default = "xtools_bash"
        
    class KeyboardAdaptor:
        default = (pynput, evdev)
        Dfuncs = pynput
        keywait = (evdev, pynput)
        hotstring = pynput
        
    class IconAdaptor:
        default = (pystray, "yad_bash")
        
    class ClipboardAdaptor:
        default = "xclip_bash"
    
    class NotificationAdaptor:
        default = "notify_send_bash"
        waiter =  default
        
    class DialogAdaptor:
        default = dict(_="zenity_bash",popup="tkinter")
        
    class ImageAdaptor:
        default = "wand"
        
    class mp3tagAdaptor:
        default ="eyed3"
        
    class TextToSpeechAdaptor:
        default = "ttsx3"
        

class termux:
    class ImageAdaptor(linux.ImageAdaptor):
        pass
    
    class mp3tagAdaptor(linux.mp3tagAdaptor):
        pass
    
    
class windows:
    class KeyboardAdaptor:
        default = pynput
        Dfuncs = pynput
        keywait = pynput
        hotstring = pynput
        
    class IconAdaptor:
        default = pystray
        
    class MouseAdaptor:
        default = pynput
        Dfuncs = default
        keywait = default

    class ImageAdaptor:
        default = "wand"

    class mp3tagAdaptor:
        default = "eyed3"


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from ... import adaptionmethod
from ..event_sender import AdaptiveSender, AdaptivePressReleaseSender


class SenderAdaptorMixin:
    
    
    default_delay = 0
    """
    .. exec::

        event_sender.default_delay_doc("events")
    """
    
    def __init__(self, *args,**kwargs):
        super().__init__(*args, **kwargs)
        self.selected_devs = []
        
    @adaptionmethod
    def select_device(self, **kwargs):
        devices = self.select_device.target(**kwargs)
        if not devices: raise ValueError
        self.selected_devs+= devices
        return self.selected_devs

class SenderAdaptor(SenderAdaptorMixin, AdaptivePressReleaseSender):
    
    # inherits adaptionmethods press and rls
    
    default_duration = 1
    """
    .. exec::

        event_sender.default_duration_doc("event object")
    """


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.baseclasses import (check_type, AdditionContainer)
from Dhelpers.arghandling import (PositiveInt, CollectionWithProps)

import inspect, time, functools
from abc import ABC, abstractmethod


class WindowNotFoundError(Exception):
    pass


def doc_single(prop):
    def dec(func):
        doc=(f"Returns the {prop} of this window if exactly one "
                "match was found. Returns ``None`` if no match was "
              "found. Raises ``ValueError`` if there was more than one "
                  "matching window.")
        # doc = (f":return: - The {prop} of this window if exactly one "
        #        "match was found. \n \t - ``None`` if no match was "
        #        "found. \n :raise ValueError: If there was more than one "
        #        "matching window.")
        try:
            add = func.__doc__
        except AttributeError:
            pass
        else:
            if add:
                #add = add.replace("\t","")
                #while "  " in add: add = add.replace("  "," ")
                doc += "\n" + add
        func.__doc__ = doc
        return func
    return dec

def doc_multi(prop, no_gen=False):
    def dec(func):
        func.__doc__=(f"Returns a tuple of length :data:`num` containing "
                      f"the {prop} of all found windows.")
        if no_gen is False:
            func.__doc__ += (f" Use :func:`{func.__name__}_gen` to create a "
                      f"generator instead of a tuple.")
        return func
    return dec




class WindowObject(ABC):
    
    @classmethod
    def cached(cls):
        return cls.adaptor.cached()
    
    @classmethod
    def cached_properties(cls):
        return cls.adaptor.cached_properties
    
    @classmethod
    def use_cache(cls, state=True):
        cls.adaptor.use_cache=state
    
    
    #to be set by subclasses:
    adaptor = None
    _FoundWinClass = None
    _WinSearchClass = None
    
    @staticmethod
    def _make_single_val(list_of_vals):
        if inspect.isgenerator(list_of_vals):
            try:
                val = list_of_vals.__next__()
            except StopIteration:
                return  # return None if no matching window found
            try:
                list_of_vals.__next__()
            except StopIteration:
                return val  # return the found val if exactly one found
        else:
            l = len(list_of_vals)
            if l == 0: return
            if l == 1: return list_of_vals[0]
        raise ValueError("More than one matching window found.")
    
    @property
    def num(self):
        """Number (*int*) of windows found to match the given criteria. Equals
            zero if there was no match at all."""
        return len(self.IDs())
    
    
    def __bool__(self):
        # This allows to use if win("test"): do_something
        return bool(self.IDs())
    
    @abstractmethod
    def existing_IDs(self):
        raise NotImplementedError
    
    @doc_single("identification number (*int*)")
    def ID(self):
        return self._make_single_val(self.IDs())
    
    @abstractmethod
    @doc_multi("IDs", no_gen=True)
    def IDs(self):
        raise NotImplementedError
    
    @doc_single("title (*str*)")
    def title(self):
        return self._make_single_val(self.titles_gen())
    @doc_multi("titles")
    def titles(self):
        return tuple(self.titles_gen())
    def titles_gen(self):
        for ID in self.IDs():
            x = self.adaptor.property_from_ID(ID, "title")
            check_type(str, x, allowed=(None,))
            yield x
    
    @doc_single("window class (*str*)")
    def wcls(self):
        return self._make_single_val(self.wclasses_gen())
    @doc_multi("window classes")
    def wclasses(self):
        return tuple(self.wclasses_gen())
    def wclasses_gen(self):
        for ID in self.IDs():
            x = self.adaptor.property_from_ID(ID, "wcls")
            check_type(str, x, allowed=(None,))
            yield x
    
    
    @doc_single("process ID (*int*)")
    def pid(self):
        return self._make_single_val(self.pids_gen())
    @doc_multi("process IDs")
    def pids(self):
        return tuple(self.pids_gen())
    def pids_gen(self):
        for ID in self.IDs():
            x = self.adaptor.property_from_ID(ID, "pid")
            check_type(PositiveInt, x, allowed=(None,))
            yield x
    
    @classmethod
    def screen_res(cls):
        """Returns the screen resolution in pixels as a
            tuple (*screen_width, screen_height*)."""
        return cls.adaptor.screen_res()
    
    def widths_gen(self):
        for geom in self.geometries_gen(): yield geom[2] - geom[0]
    @doc_single("width in pixels (*int*)")
    def width(self):
        return self._make_single_val(self.widths_gen())
    @doc_multi("widths")
    def widths(self):
        return tuple(self.widths_gen())
    
    
    @doc_single("height in pixels (*int*)")
    def height(self):
        return self._make_single_val(self.widths_gen())
    def heights_gen(self):
        for geom in self.geometries_gen(): yield geom[3] - geom[1]
    @doc_multi("heights")
    def heights(self):
        return tuple(self.widths_gen())
    
    
    @doc_single("4 border coordinates")
    def geometry(self):
        """
        The return value is a 4-element tuple containing the window's
        border coordinates as *int* in pixels as follows: (left, top,
        right, bottom) = (Xmin, Ymin, Xmax, Ymax).
        """
        return self._make_single_val(self.geometries_gen())
    def geometries_gen(self):
        for ID in self.IDs():
            x = self.adaptor.property_from_ID(ID, "geometry")
            check_type(CollectionWithProps(int, len=4), x)
            yield x
    @doc_multi("geometries")
    def geometries(self):
        return tuple(self.geometries_gen())
    
    
    @doc_single("2-element tuple (:func:`title`, :func:`wcls`)")
    def info(self):
        return self._make_single_val(self.infos_gen())
    def infos_gen(self):
        for a in zip(self.titles_gen(), self.wclasses_gen()): yield a
    @doc_multi(":func:`info` tuples")
    def infos(self):
        return tuple(self.infos_gen())
    
    
    
    def all_infos_gen(self):
        for a in zip(self.IDs(), self.titles_gen(), self.wclasses_gen(),
                self.pids_gen(), self.geometries_gen()): yield a
    @doc_single("5-element tuple (:func:`ID`, :func:`title`, :func:`wcls`, "
                ":func:`pid`, :func:`geometry`)")
    def all_info(self):
        return self._make_single_val(self.all_infos_gen())
    @doc_multi(":func:`all_info` tuples")
    def all_infos(self):
        return tuple(self.all_infos_gen())
    
    def print_all_infos(self):
        print(*self.all_infos(), sep="\n")
        return self
    
    def close(self, all=False):
        """Closes this window.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
           window matches.
        """
        if all:
            for ID in self.IDs(): self.adaptor.close(ID)
        else:
            return self.adaptor.close(self.ID())
        return self
    
    
    def map(self, all=False):
        if all:
            for ID in self.IDs(): self.adaptor.map(ID)
        else:
            return self.adaptor.map(self.ID())
        return self
    
    def set_prop(self, action: str, prop: str, prop2: str = False, all=False):
        if all:
            for ID in self.IDs():
                self.adaptor.set_prop(ID, action, prop, prop2)
        else:
            return self.adaptor.set_prop(self.ID(), action, prop, prop2)
        return self
    
    
    def kill(self, all=False):
        """Kill (force close) this window.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than
            one  window matches.
        """
        if all:
            for ID in self.IDs(): self.adaptor.kill(ID)
        else:
            return self.adaptor.kill(self.ID())
        return self
    
    def minimize(self, all=False):
        """Minimize this window.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
         window matches.
        """
        if all:
            for ID in self.IDs(): self.adaptor.minimize(ID)
        else:
            return self.adaptor.minimize(self.ID())
        return self
    
    def maximize(self, all=False):
        """Maximize this window.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
         window matches."""
        sr = self.adaptor.screen_res()
        self.move(0, 0, *sr, all=all)
        return self
    
    def max_left(self, all=False):
        """Maximize this window to the left side of the screen.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
         window matches."""
        try:
            self.set_prop("remove", "maximized_horz", prop2="maximized_vert")
        except Exception:
            pass
        a, b = self.adaptor.screen_res()
        self.move(0, 0, a/2, b, all=all)
        return self
    
    def max_right(self, all=False):
        """Maximize this window to the right side of the screen.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
         window matches."""
        try:
            self.set_prop("remove", "maximized_horz", prop2="maximized_vert")
        except Exception:
            pass
        a, b = self.adaptor.screen_res()
        self.move(a/2, 0, a/2, b, all=all)
        return self
    
    def move(self, x=-1, y=-1, width=-1, height=-1, all=False):
        """Move and resize this window. Specifying -1 for one of the
        coordinates will keep the current value (default).

        :param int x: New coordinate of left border.
        :param int y: New cordinate of top border
        :param int width: New width of window.
        :param int height: New hight of window.
        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
         window matches.
        """
        if all:
            for ID in self.IDs(): self.adaptor.move(ID, x, y, width, height)
        else:
            return self.adaptor.move(self.ID(), x, y, width, height)
        return self
    
    def activate(self, all=False):
        """Activates this window.

        :param all: If set ``True``, apply this command to all found windows.
        :raise ValueError: If *all* is ``False`` (default) and more than one
         window matches.
        """
        if all:
            for ID in self.IDs(): self.adaptor.activate(ID)
        else:
            return self.adaptor.activate(self.ID())
        return self
    
    def wait_active(self, timeout=5, pause_when_found=0.05, timestep=0.2,
            reverse=False):
        """Wait until the specified window is active. If several windows
        are matching, wait until any of them is active.

        :param float timeout: Seconds to wait before returning anyway.
        :param float pause_when_found: Time to sleep after success.
        :param float timestep: Interval how often the condition is checked.
        :return: - The :func:`Win` object for the active window in case of success.
                - ``False`` in case of a timeout.

        """
        waited = 0
        while waited <= timeout:
            awin = self._FoundWinClass()
            # this WindowSearch Object is refreshed everytime
            if awin.ID() in self.IDs():
                if not reverse:
                    time.sleep(pause_when_found)
                    return awin
            elif reverse:
                return True
            if timeout == 0:
                break  # zero length timeout: if not matched go straight to end
            time.sleep(timestep)
            waited += timestep
        return False
    
    def wait_not_active(self, timeout=5, timestep=0.2):
        """Same as :func:`wait_active`, but reversed. It waits until the
        specified window(s) is (are) not active anymore.

        :return: - ``True`` in case of success.
                - ``False`` in case of a timeout.
        """
        return self.wait_active(timeout=timeout, timestep=timestep,
                reverse=True)
    
    
    def wait_exist(self, timeout=5, pause_when_found=0.05, timestep=0.2,
            min_wincount=1, max_wincount=None):
        """Waits until the number of matching windows is between
        *min_wincount* and *max_wincount*. By default, this means at least
        one matching window must exist.

        :param float timeout: Seconds to wait before returning anyway.
        :param float pause_when_found: Time to sleep after success.
        :param float timestep: Interval how often the condition is checked.
        :param int min_wincount: Minimum number of matching windows
            necessary to continue.
        :param max_wincount: Maximum number of matching windows to continue.
            If set to ``None`` (default), there is no maximum.
        :return: - The :func:`Win` object containing all matching windows in case of success.
             - ``False`` in case of a timeout.
        """
        waited = 0
        while waited <= timeout:
            IDs = self.existing_IDs()
            if len(IDs) >= min_wincount:
                if max_wincount is None or len(IDs) <= max_wincount:
                    # if appropriate number of existing windows was found,
                    # return the IDs
                    time.sleep(pause_when_found)
                    if IDs: return self._FoundWinClass(IDs, at_least_one=True)
                    return True
            if timeout == 0:
                break  # zero length timeout, if not matched go straight to end
            time.sleep(timestep)
            waited += timestep
        return False
    
    
    def wait_not_exist(self, timeout=5, timestep=0.2):
        """Wrapper for :func:`wait_exist` with *min_wincount* =
        *max_wincount* = 0.

        :return: - ``True`` in case of success.
                - ``False`` in case of a timeout."""
        ret = self.wait_exist(timeout, 0, timestep, min_wincount=0,
                max_wincount=0)
        check_type(bool, ret)
        return ret
    

    def wait_num_change(self, num_change, timeout=5, pause_when_found=0.05,
            timestep=0.2):
        """Wrapper for :func:`wait_exist`. Waits until the number of
        currently matching windows has changed by *num_change* (relative to
        the initial found number :data:`num`).

        :param int num_change: A positive or negative integer.
        :return: - The :func:`Win` object of the newly found window(s) if *num_change* is positive.
            - The :func:`Win` object of the closed window(s) if *num_change* is negative. (These do not exist anymore.)
            - ``False`` in case of a timeout.
        
        Useful when opening a new window if other windows of the same class
        are already existing, see for example:
        :ref:`Launch the browser and simultaneously redirect
        key presses`.
        """
        raise NotImplementedError
    
    def wait_exist_activate(self, timeout=5, timestep=0.2):
        """Wait until at least one matching window exists and activate it immediately.

        :return: - The :func:`Win` object for the found window.
            - ``False`` in case of timeout or if activate failed.
        """
        found_win = self.wait_exist(timeout=timeout, timestep=timestep,
                pause_when_found=0)
        if found_win:
            found_win.activate()
            return found_win.wait_active(timeout=1)
        return False


class WindowSearch(AdditionContainer.Addend, WindowObject):
    
    # _FoundWinClass and _WinSearchClass are set by
    # init_subclass of FoundWindows
    
    @property
    def adaptor(self):
        return self._FoundWinClass.adaptor
    
    def find(self):
        return self._FoundWinClass(self)
    
    def update_properties(self, **kwargs):
        new_kwargs = self.creation_kwargs.copy()
        new_kwargs.update(kwargs)
        return self.__class__(*self.creation_args, **new_kwargs)
    
    def process_args(self, title_or_ID=None, wcls = None, *, loc=None,
            limit=None, visible=None, selection_func=None, **properties):
        if wcls is not None: properties["wcls"] = wcls
        check_type(int, limit, allowed=(None,))
        self.visible = visible
        self.location = loc
        self.fixed_IDs = None
        if selection_func:
            assert callable(selection_func)
            raise ValueError("selection_func not yet implemented")
        self.selection_func = selection_func
        if loc is not None:
            if properties or title_or_ID or limit:
                raise SyntaxError("Parameter loc cannot be combined with "
                                  "other parameters.")
        else:
            self.limit = limit
            if isinstance(title_or_ID, PositiveInt):
                self.fixed_IDs = (title_or_ID,)
            elif isinstance(title_or_ID, CollectionWithProps(PositiveInt)):
                self.fixed_IDs = tuple(title_or_ID)
            elif isinstance(title_or_ID, (str, CollectionWithProps(str))):
                properties["title"] = title_or_ID
            elif not title_or_ID:
                if not properties:
                    self.location = "active"
                    # if no parameters specified, use the active window
            else:
                raise TypeError("First parameter title_od_ID: %s\nhas wrong "
                                "type: %s"%(title_or_ID, type(title_or_ID)))
            self._properties = properties

    @functools.wraps(process_args)
    def __init__(self, *winargs, **winkwargs):
        self.creation_args = winargs
        self.creation_kwargs = winkwargs
        self.process_args(*winargs, **winkwargs)
    
    def compare_args(self, *winargs,**winkwargs):
        new_instance = self.__class__(*winargs, **winkwargs)
        for attr in ("location", "fixed_IDs"):
            if getattr(self,attr) != getattr(new_instance, attr):
                return False
        for key, val in self._properties.items():
            newval = new_instance._properties.get(key)
            if isinstance(val, CollectionWithProps()):
                val = set(val)
                newval = set(newval)
            if newval != val: return False
        return True
    
    def IDs(self):
        if self.location:
            return (self.adaptor._ID_from_location(self.location),)
        
        winlist = None
        if self.fixed_IDs:
            winlist = set(ID for ID in self.fixed_IDs if
                self.adaptor.id_exists(ID))
        
        for prop, prop_val in self._properties.items():
            if isinstance(prop_val, CollectionWithProps()):
                matching_ids = set()
                for val in prop_val:
                    matching_ids |=  set(self.adaptor.IDs_from_property(prop,
                            val, visible=self.visible))
            else:
                matching_ids = set(self.adaptor.IDs_from_property(prop,
                        prop_val, visible=self.visible))
            if winlist is None:
                winlist = matching_ids
            else:
                winlist &= matching_ids  # make intersection
            if winlist is set(): break
        out = tuple(sorted(winlist))
        l = self.limit
        if l:
            if l > 0:
                out = out[:l]
            elif l < 0:
                out = out[l:]
            else:
                raise ValueError(self.limit)
        # if self.selection_func:
        #     out = tuple(ID for ID in out if self.selection_func(
        #             self._FoundWinClass(ID)))
        return out
    
    
    def check_single(self, win):
        # problem here: we don't know how to check whether win is visible..
        # so self.visible has no effect inside here.
        if isinstance(win, int):
            id = win
        elif isinstance(win,self._FoundWinClass):
            id = win.ID()
        else:
            raise TypeError
        if self.location: return id in self.IDs()
        # in other cases we don't want to run self.IDs() because it scans for
        # all possible matching IDs taking more time
        if self.fixed_IDs and id not in self.fixed_IDs: return False
        for prop, val in self._properties.items():
            winval = self.adaptor.property_from_ID(id, prop)
            if isinstance(val,CollectionWithProps()):
                if winval not in val: return False
            elif winval != val:
                return False
        return True
    
    def existing_IDs(self):
        return self.IDs()
    
    
    
    
    # the following redefinitions are necessary because this way,
    # the self.IDs() search will only be executed once.
    def infos_gen(self):
        return WindowObject.infos_gen(self.find())
    
    def all_infos_gen(self):
        return WindowObject.all_infos_gen(self.find())
    
    def wait_num_change(self, *args, **kwargs):
        return self.find().wait_num_change(*args,**kwargs)

    



class WindowSearchContainer(AdditionContainer, WindowSearch,
        basic_class = WindowSearch):
    
    def IDs(self):
        ids = set()
        for winsearch in self.members: ids |= set(winsearch.IDs())
        # create the union of all found window IDs
        return tuple(sorted(ids))
    
    @property
    def _FoundWinClass(self):
        return self.members[0]._FoundWinClass
    
    def compare_args(self, *winargs,**winkwargs):
        for member in self.members:
            if member.compare_args(*winargs, **winkwargs): return True
        return False

    def check_single(self, win):
        for member in self.members:
            if member.check_single(win): return True
        return False


class FoundWindows(WindowObject):
    """An object of this class represents one or more graphical windows of
    your operating system.
    
    When a new instance of this class is created, a search is
    performed to find all existing windows matching the given parameters.
    Each window is identified by a unique system-wide identification number (ID).
    The matching IDs are saved in ascending order as a static list. Use
    the :func:`IDs` method to see them. All methods below will operate on
    this list of initially found windows.
     
    Two :func:`Win` objects are considered identical if their list
    of :func:`IDs` are identical.
    
    """
    
    def __init_subclass__(cls):
        class Search(WindowSearch):
            pass
        Search.__module__ = __name__
        Search._FoundWinClass = cls
        Search._WinSearchClass = Search
        # winsearch.__init__=functools.wraps(WindowSearch.__init__)(winsearch.__init__)
        cls._FoundWinClass = cls
        cls._WinSearchClass = Search
        cls.Search = Search
        cls._search_activewin = Search()
        base_doc = cls.__base__.__doc__
        if base_doc:
            this_doc = cls.__doc__
            if not this_doc:
                new_doc = base_doc
            else:
                new_doc = this_doc + "\n\n" + base_doc
            new_doc = new_doc.replace("\n  ", "\n")
            cls.__doc__ = new_doc
            
            
    def _signature_def(self, title_or_ID =None, wcls = None, *, pid = None,
            loc = None, at_least_one = False, limit = None, visible = None):
        return
    
    def __init__(self, *winargs, at_least_one = False, **winkwargs):
        """
        :param title_or_ID:
            - A string is interpreted as the title to search for. The string
              can be contained anywhere in the window's title to
              be considered a match.
            - An integer is interpreted as a window ID. Specifying this
              directly will skip the search and all other parameters are hence
              ignored.
        :type title_or_ID: str, int
        :param str wcls: Window class to search for. This must be an exact
                match.
        :param int pid: Process ID of the window to search for.
        :param tuple loc: Specify screen coordinates (x,y) for this
            parameter to retrieve the top-most window at that location.
            Specifying this will ignore all other parameters.
        :param bool at_least_one: If set to ``True``, check that at least one
            matching window has been found.
        :param int limit: Maximum number of windows to include into the
            window object. If more matching windows exist, only the ones with
            the lowest IDs will be included.
        :param bool visible: If set to True, only visible windows will be
            considered.
        :raise WindowNotFoundError: If *at_least_one* is ``True`` and no
            matching window was found.
        
        In order to find the currently active (i.e. topmost) window,
        simply call this class without parameters::
            
            active_window = Win()
            
        Several different parameters can be
        passed at once to narrow down the search. A window must
        fullfill all specified parameters to be a match::
            
            dpowers_doc_win = Win('Dpowers documentation', 'Navigator.firefox')
            # searches for windows with
            # 'Dpowers documentation' contained in title
            # AND
            # window class is equal to 'Navigator.firefox' (Firefox browser)
            
        In place of passing a single value for the parameters *ID*, *title*,
        *wcls* or *pid*, it is possible to pass a
        list (or tuple or set) of allowed values instead, which will search
        for a wider scope of windows::
        
            dpowers_browser_win = Win('Dpowers documentation',
                ('Navigator.firefox','chromium-browser.Chromium-browser'))
            # searches for windows with
            # 'Dpowers documentation' contained in title
            #   AND
            # (window class is equal to 'Navigator.firefox'
            #  OR window class is equal to 'chromium-browser.Chromium-browser')

        In many cases it is desirable to make sure that exactly one
        matching window is found for each instance to avoid
        confusion. In other cases it might be useful to operate on a
        group of windows simultaneously.
        
        
        .. note:: If a window was initially found and closed in the meantime,
            its ID will still be in the internal list of the :func:`Win` object.
            This can result in an Exception when trying to interact with a
            non-existent window ID. The :func:`update` or
            :func:`remove_non_existing` methods can be used to avoid this
            problem.

        """
        if len(winargs) == 1 and not winkwargs and isinstance(winargs[0], WindowSearch):
            WinSearch_instance = winargs[0]
            if not isinstance(WinSearch_instance, WindowSearchContainer):
                if WinSearch_instance.adaptor is not self.adaptor:
                    raise TypeError
            self.winsearch_object = WinSearch_instance
            # reuse the WindwoSearch object
        elif not winargs and not winkwargs:
            self.winsearch_object = self._search_activewin
        else:
            self.winsearch_object = self._WinSearchClass(*winargs, **winkwargs)
            # this will initialize instance attributes according to
            # WindowSearch init method.
        self.found_IDs = self.winsearch_object.IDs()
        if not self.found_IDs:
            self.found_IDs = ()
            if at_least_one: raise WindowNotFoundError(
                "\nCould not find windows of specified properties. Please "
                "use WindowSearch class instead.")
    
    
    __init__.__signature__ = inspect.signature(_signature_def)
    
    def IDs(self):
        # careful: these IDs might not be existing any more
        return self.found_IDs
    
            
    
    def check(self, *winargs, **winkwargs):
        assert len(self) == 1
        first = winargs[0]
        id = self.ID()
        if isinstance(first, WindowSearch):
            assert not winkwargs
            for obj in winargs:
                if obj.check_single(id): return True
            return False
        search = self._WinSearchClass(*winargs,**winkwargs)
        return search.check_single(id)
    
    def update(self):
        """Re-perform the initial search for matching existing windows and
        update this window object's internal ID list. This removes
        non-existing windows and searches for new matches.
        """
        self.found_IDs = self.winsearch_object.IDs()
        
    def update_parameter(self,**kwargs):
        new_winsearch = self.winsearch_object.update_properties(**kwargs)
        return new_winsearch.find()
    
    def find(self):
        return self.winsearch_object.find()
    
    def existing_IDs(self):
        return tuple(ID for ID in self.IDs() if self.adaptor.id_exists(ID))
    
    def remove_non_existing(self):
        """Cleanse this window object's internal ID list by removing those
        that do not exist anymore."""
        self.found_IDs = self.existing_IDs()
    
    
    def wait_active(self, timeout=5, pause_when_found=0.05, timestep=0.2):
        if self.found_IDs:
            return super().wait_active(timeout, pause_when_found, timestep)
        return self.winsearch_object.wait_active(timeout, pause_when_found, timestep)
    
    def wait_exist(self, timeout=5, pause_when_found=0.05, timestep=0.2,
            min_wincount=1, max_wincount=None):
        if self.num >= min_wincount:
            # this means that this instane of WindowFinder contains
            # enough window IDs to theoretically satisfy the condition
            return super().wait_exist(timeout, pause_when_found, timestep,
                    min_wincount, max_wincount)
            # raise ValueError("WindowFinder does not contain enough IDs "
            #                "to satisfy condition.")
        # if this FoundWindow instance does not contain enough windows,
        # then just use the windowsearch used to define this instance
        # this way we can write win("title").wait_exist() even if the title
        # does not yet exist. Instead fo win.Search("title").wait_exist()
        return self.winsearch_object.wait_active(timeout, pause_when_found, timestep)
    
    
    def wait_num_change(self, num_change, timeout=5, pause_when_found=0.05,
            timestep=0.2):
        check_type(int, num_change)
        kwargs = {}
        if num_change >= 0:
            kwargs["min_wincount"] = self.num + num_change
        if num_change <= 0:
            kwargs["max_wincount"] = self.num + num_change
        windows_after = self.winsearch_object.wait_exist(timeout, pause_when_found,
                timestep, **kwargs)
        if windows_after is not False:
            if num_change >= 0:
                return windows_after - self
            return self - windows_after
        return False
    
    
    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self._FoundWinClass(set(self.IDs()) | set(other.IDs()))
        return NotImplemented
    __or__ = __add__
    
    
    def __sub__(self, other):
        if isinstance(other, self.__class__):
            return self._FoundWinClass(set(self.IDs()) - set(other.IDs()))
        return NotImplemented
    # def __rsub__(self, other):
    #     if isinstance(other, self.__class__):
    #         return other.__sub__(self)
    #     return NotImplemented
    
    def __and__(self, other):
        if isinstance(other, self.__class__):
            return self._FoundWinClass(set(self.IDs()) & set(other.IDs()))
        return NotImplemented
    
    
    def __repr__(self):
        return "%s with window ID(s) %s>"%(super().__repr__()[:-1], self.IDs())


    def __getitem__(self, item):
        return self.__class__(self.IDs()[item])
    
    def __iter__(self):
        def iterator():
            for id in self.IDs(): yield self.__class__(id)
        return iterator()
    
    def __len__(self):
        return len(self.IDs())


    def __eq__(self, other):
        if isinstance(other, WindowObject):
            return set(self.IDs()) == set(other.IDs())
        if isinstance(other, PositiveInt): return other == self.ID()
        if isinstance(other, str): return self.title() == other
        return NotImplemented

    def __hash__(self):
        if self.num == 0: return 0
        return max(self.IDs())

    def __contains__(self, item):
        if isinstance(item, PositiveInt):
            return item in self.IDs()
        if isinstance(item, WindowObject):
            return item.ID() in self.IDs()
        if isinstance(item, str):
            for t in self.titles_gen():
                if item in t: return True
            return False
        return NotImplemented



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from warnings import warn
from Dhelpers.baseclasses import AdditionContainer, TimedObject
from Dhelpers.container import container
from Dhelpers.arghandling import check_type
from abc import ABC, abstractmethod
import threading, queue, logging, inspect, functools, traceback
from ..event_classes import MouseMoveEvent
from Dhelpers.named import NamedObj

class CallbackRunnerThread(threading.Thread):
    pass

class EventQueueThread(threading.Thread):
    pass


class EventQueueReader:
    def __init__(self):
        # create useless dead thread just to make all the attributes available:
        self.callback_runner_thread = EventQueueThread()
        self.event_queue = queue.Queue()
    
    def start(self):
        if not self.callback_runner_thread.is_alive():
            self.callback_runner_thread = EventQueueThread(
                    target=self.run_callbacks)
            self.callback_runner_thread.start()
            return self.callback_runner_thread
    
    def run_callbacks(self):
        while True:
            handler, event_obj = self.event_queue.get()
            handler.run_callbacks(event_obj)




class InputEventHandler(ABC):
    
    # base class to create Capturer and Collector classes
    
    reinject_implemented = False
    capture_allowed = True  #need to be manually activated
    collect_allowed = True  #need to be manually activated
    
    def __init__(self, hook_cls=None):
        self.hook_cls = hook_cls  #this attribute can be left None as it will
        # usually be defined by __init_subclass__ method of CallbackHook class
        self.collect_active = False
        self.capture_active = False
        self.active_hooks = []
        self.cap_num = 0


    def add_hook(self, hook):
        ah = self.active_hooks
        ah.append(hook)
        ah.sort(key=lambda s: s.priority)
        if len(ah) == 1:
            if self.collect_active: raise RuntimeError
            self.collect_active = True
            self.start_collecting()
            return True

    def remove_hook(self, hook):
        self.active_hooks.remove(hook)
        if len(self.active_hooks) == 0:
            if not self.collect_active: raise RuntimeError
            self.collect_active = False
            self.stop_collecting()
            return True
    
    def add_capturer(self):
        self.cap_num += 1
        if self.cap_num == 1:
            if self.capture_active: raise RuntimeError
            self.capture_active = True
            self.start_capturing()
            return True

    def remove_capturer(self):
        self.cap_num -= 1
        if self.cap_num == 0:
            if not self.capture_active: raise RuntimeError
            self.capture_active = False
            self.stop_capturing()
            return True
        
        
    def start_collecting(self):
        # must be non_blocking
        raise NotImplementedError
        
    def stop_collecting(self):
        pass


    def start_capturing(self):
        # must be non_blocking
        raise NotImplementedError

    def stop_capturing(self):
        pass
        
    
    def queue_event(self, *args, **kwargs):
        try:
            cls = self.hook_cls
            event_obj = cls._create_event_obj(*args, **kwargs)
            if cls.reinject_active:
                ret = cls._active_reinject_func(event_obj)
                #print(ret, cls._active_reinject_func)
                if ret is None: raise ValueError
            else:
                ret = None
            cls.queue_reader.event_queue.put((self,event_obj))
            return ret
        except Exception as e:
            warn(f"\nEmergency stop {self.hook_cls}.\n{repr(e)}")
            try:
                self.hook_cls.capturer.stop()
            except Exception:
                pass
            try:
                self.stop_collecting()
            except Exception:
                pass
            try:
                self.stop_capturing()
            except Exception:
                pass
            traceback.print_exc()
    
    def run_callbacks(self, event_obj):
        if event_obj is None: return
        for hook in self.active_hooks:
            ret = hook.run_callback(event_obj)
            if ret == "block": break
            
            
    
    

class Missing: pass
    
    
class CallbackHook(AdditionContainer.Addend, TimedObject):
    
    # usually set by __init_subclass__ or subclass attribute. In these cases,
    # there is at most one handler instance for each subclass.
    # They can optionally be set for a particular Hook instance, in which
    # case this instance has it's own handler available (CustomHook subclass
    # uses this for evdev)
    handler = None
    
    @property
    def capture_allowed(self):
        return self.handler.capture_allowed
        
    @property
    def collect_allowed(self):
        return self.handler.collect_allowed
    
    
    queue_reader = EventQueueReader()
    # this should usually never be overwritten as all subclasses
    # should use a single Thread for executing the callbacks. If a
    # subclass defines a seperate instance of EventQueueReader, its
    # callbacks will be run in a separate thread
    
    def __init_subclass__(cls):
        cls._active_capturers = 0
        cls._active_reinject_func = None
        cls.reinject_active = False
        handler = cls.handler
        if handler is not None:
            check_type(InputEventHandler, handler)
            if handler.hook_cls is None: handler.hook_cls = cls
            if handler.hook_cls is not cls: raise ValueError
        cls.__init__.__signature__ = inspect.signature(cls.init)



    def __init__(self, *args, **kwargs):
        self._init_kwargs = self._swallow_args(*args,**kwargs)
        self.init(**self._init_kwargs)
    
    @staticmethod
    def _swallow_args(callback_func=Missing,timeout=Missing, **kwargs):
        # btw: makes sure that no positional arg is also specified in kwargs
        if callback_func is not Missing: kwargs["callback_func"]=callback_func
        if timeout is not Missing: kwargs["timeout"] = timeout
        return kwargs
    
    def init(self, callback_func = False, timeout = 60, *, capture: bool =
            False, reinject_func = None, priority: int  = 0,
            dedicated_thread=False, **custom_kwargs):
        super().__init__(timeout=timeout)
            # this defines the attributes timeout, active
        if callback_func is None:
            callback_func = lambda *args: None
        elif not callable(callback_func) and callback_func is not False:
            raise TypeError("callback_func must be callable, False or None")
        else:
            argcount = callback_func
        check_type(int, priority)
        check_type(bool, capture)
        self.callback_func = callback_func
        self.priority = priority
        self.reinject_func = reinject_func
        self.capture = capture
        self.custom_kwargs = custom_kwargs
        self.dedicated_thread = dedicated_thread
        self.process_custom_kwargs(**custom_kwargs)
        if self.callback_func and not self.collect_allowed: raise ValueError
        if reinject_func is not None:
            if self.handler.reinject_implemented is False:
                raise NotImplementedError(
                    "The following hook implementation does not support "
                    "reinjecting:\n" + str(self.__class__))
            if callback_func is False:
                self.callback_func = lambda *args: None
            self.capture = True
            if not callable(reinject_func):
                self.reinject_func = lambda *args: reinject_func
        if self.capture is True and not self.capture_allowed:
            raise NotImplementedError(
                    "The following hook implementation does not support "
                    "capturing:\n" + str(self.__class__))
    
    def __call__(self, *args, _reuse=True, **kwargs):
        """Create a copy, Adding left out arguments.
        For a useful example of successively adding argument see
        Dpowers.events.waiter.Keywaiter class."""
        if not _reuse: return self.__class__(*args,**kwargs)
        newkwargs = self._init_kwargs.copy()
            #this includes initial *args and **kwargs values
        newkwargs.update(self._swallow_args(*args, **kwargs))
            #this will overwrite inital *args and **kwargs value
        return self.__class__(**newkwargs)

    def __bool__(self):
        return self.callback_func is not False or self.capture is True

        
    def process_custom_kwargs(self, **kwargs):
        if not kwargs: return
        raise NotImplementedError(f"Hook {self} does not accept custom kwargs.")


    # this method is called from cls.collector instance to run the callbacks
    def run_callback(self, event_obj):
        if self.callback_func is None: return
        event_obj = self.callback_condition(event_obj)
        if event_obj is None: return
        # logging.info("calling function %s with argument %s.",
        #       self.callback_func, event_obj)
        if self.dedicated_thread:
            CallbackRunnerThread(target=self.callback_func,
                    args=(event_obj,)).start()
        else:
            ret = self.callback_func(event_obj)
            # this is where the callback is actually executed. Can
            # take some while
            if ret == "stop": self.stop()
            return ret
    
    #start method is defined in TimedObj Baseclass and will use this:
    def _start_action(self):
        cls = self.__class__
        if self.reinject_func is not None:
            if cls._active_reinject_func is None:
                cls._active_reinject_func = self.reinject_func
            else:
                raise RuntimeError("Multiple reinject functions set for the "
                        "same hook class:\n" + str(cls))
        if self.capture:
            cls._active_capturers += 1
            cls._update_reinject()
            self.handler.add_capturer()
        if self.callback_func is not False:
            self.queue_reader.start()  # just in case the queue_reader has
            # somehow been stopped, lets recheck everytime
            self.handler.add_hook(self)

    # stop method is defined in TimedObj Baseclass and will use this:
    def _stop_action(self):
        cls = self.__class__
        if self.reinject_func is not None:
            cls._active_reinject_func = None
        if self.capture:
            cls._active_capturers -= 1
            cls._update_reinject()
            self.handler.remove_capturer()
        if self.callback_func is not False:
            self.handler.remove_hook(self)
    
    def _timeout_action(self):
        warn("Timeout after %s seconds: Stopping hook %s." % (
            self.timeout,self))
        
    @classmethod
    def _update_reinject(cls):
        cls.reinject_active = cls._active_reinject_func is not None and \
                              cls._active_capturers == 1


    # @classmethod
    # def queue_event(cls, *args, **kwargs):
    #     event_obj = cls._create_event_obj(*args, **kwargs)
    #     if cls.reinject_active:
    #         ret = cls._active_reinject_func(event_obj)
    #         if ret is None: raise ValueError
    #     else:
    #         ret = None
    #     cls.queue_reader.event_queue.put((cls, event_obj))
    #     return ret

    

    # @classmethod
    # def _run_callbacks(cls, event_obj):
    #     if event_obj is None: return
    #     for self in cls.active_callbacks:
    #         if self.callback_func is None or not self.callback_condition(
    #                 event_obj): continue
    #         logging.info("calling function %s with argument %s.",
    #                 self.callback_func, event_obj)
    #         ret = self.callback_func(event_obj)
    #         # this is where the callback is actually executed. Can
    #         # take some while.
    #         if ret == "block": break
    #         if ret == "stop": self.stop()


    # the following method is called by the cls.collector instance to create
    # event objects.
    @classmethod
    @abstractmethod
    def _create_event_obj(cls, *args, **kwargs):
        raise NotImplementedError


    def callback_condition(self, event_obj):
        # can be overridden in subclass.
        return event_obj


class NamedHook(CallbackHook):
    NamedClass = None   #set by target_modifier of HookAdaptor
    EventClass = None
    name_translation_dicts = None
    #name_translation_dict = None
    _active_dict = {}
    
    
    
    def __init_subclass__(cls):
        super().__init_subclass__()
        if hasattr(cls, "name_translation_dict"):
            if len(cls.name_translation_dicts) != 0: raise AttributeError
            cls.name_translation_dicts = [cls.name_translation_dict]
        else:
            cls.name_translation_dicts = []

    @classmethod
    def update_active_dict(cls):
        # this has to be called after cls.NamedClass has been set
        for i in range(len(cls.name_translation_dicts)):
            dic = cls.name_translation_dicts[i]
            StandardizingDict = cls.NamedClass.StandardizingDict
            if not isinstance(dic, StandardizingDict):
                check_type(dict, dic)
                cls.name_translation_dicts[i] = StandardizingDict(dic)
        coll = cls.handler
        if coll is None: return
        check_type(InputEventHandler, coll)
        try:
            final_dict = coll.name_dict.copy()
        except AttributeError:
            final_dict = {}
        if hasattr(coll, "names"):
            for n in coll.names:
                if n not in final_dict: final_dict[n] = n
    
        for a, b in final_dict.items():
            for dic in cls.name_translation_dicts:
                try:
                    b = dic[b]
                except KeyError:
                    pass
            try:
                final_dict[a] = cls.NamedClass.instance(b)
            except KeyError:
                final_dict[a] = b
        cls._active_dict = final_dict  # this is used each time a key is
        # pressed. It  # is not standardizing to save time.
        
        
    @classmethod
    def _create_event_obj(cls, name, **kwargs):
        effective = cls._active_dict.get(name, name)
        NamedCl = cls.NamedClass
        if NamedCl is None:
            return cls.EventClass(effective, **kwargs)
        else:
            try:
                EventCl = NamedCl.Event
            except AttributeError:
                EventCl = cls.EventClass
            else:
                if cls.EventClass is not None: assert cls.EventClass is EventCl
                if isinstance(effective, NamedCl) and hasattr(effective,
                        "get_event"):
                    return effective.get_event(**kwargs)
            event_obj = EventCl(effective, **kwargs)
            # print(event_obj, event_obj.named_instance, event_obj.press)
            if not event_obj.named_instance:
                if not (event_obj.startswith("[")):
                    warn("hook of type" + str(
                            cls) + "got unknown key/button name:  " + event_obj)
                    if not event_obj.endswith("_rls"):
                        i = event_obj.lower()
                        container.set_temp_store_key("F12_send",
                                i + " = NamedKey('" + i + "')")
            return event_obj



class PressReleaseHook(NamedHook):
    
    currently_pressed = None
    
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.currently_pressed = set()
        

    def init(self, *args, press=True, release = True, write_rls=True, **kwargs):
        super().init(*args, **kwargs)
        self.press = press
        self.release = release
        self.multipress = True
        self.write_rls=write_rls
        
    
    def _stop_action(self):
        super()._stop_action()
        if not self.handler.active_hooks:
            self.__class__.currently_pressed = set()


    @classmethod
    def _create_event_obj(cls, name, press, **kwargs):
        return super()._create_event_obj(name,press=press, **kwargs)
    
    
    def callback_condition(self, event_obj):
        if event_obj.press is True:
            if not self.multipress and event_obj.multipress: return
            if self.press: return event_obj
        elif event_obj.press is False:
            if self.release:
                if self.write_rls: return event_obj
                return event_obj.strip_rls()
        else:
            return event_obj  #this happens for scroll events



class KeyhookBase(PressReleaseHook):
    
    
    def init(self, *args, allow_multipress=False, **kwargs):
        super().init(*args, **kwargs)
        if allow_multipress and not self.press: raise ValueError
        self.multipress = allow_multipress
        
    @classmethod
    def _create_event_obj(cls, name, *args, press, **kwargs):
        event_obj = super()._create_event_obj(name, *args, press=press,
                **kwargs)
        info=event_obj.name
        if event_obj.press is True:
            if info in cls.currently_pressed:
                event_obj.multipress = True
            else:
                event_obj.multipress = False
                cls.currently_pressed.add(info)
        else:
            try:
                cls.currently_pressed.remove(info)
            except KeyError:
                pass
        return event_obj


class ButtonhookBase(PressReleaseHook):
    pass


class CursorhookBase(CallbackHook):
    @classmethod
    def _create_event_obj(cls, *args,**kwargs):
        return MouseMoveEvent(*args, **kwargs)


class DefaulthookBase(CallbackHook):
    
    @classmethod
    def _create_event_obj(cls, *args):
        return args




@CallbackHook.register  #ABC classes allow registering as virtual subclass
class HookContainer(AdditionContainer, basic_class=CallbackHook, ordered=False):
    """
    This class shall synchronize the use of several hooks if you add them.
    """
    
    def __init__(self, *args):
        super().__init__(*args)
        self._members = list(m() for m in self._members)
        #this will creat a copy of all hook instances so that the
        # HookContainer Version is independent from the basic version

    def start(self):
        return tuple(m.start() for m in self._members)
    
    def stop(self):
        errors = []
        for m in self._members:
            try:
                m.stop()
            except RuntimeError as e:
                errors += [e]
        if errors:
            raise RuntimeError(str(errors))
        
    
    def join(self):
        return tuple(m.join() for m in self._members)
    
    @functools.wraps(CallbackHook.__call__)
    def __call__(self, *args, **kwargs):
        return self.__class__(*tuple(m(*args, **kwargs) for m in self._members))


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import DependencyManager, launch

with DependencyManager(__name__) as tester:
    tester.test_shellcommand("notify-send")
    

def post_notification(title,timeout,text):
    text = text.replace("<", "&lt;").replace(">","&gt;")
    # necessary because otherwise < and > will be interpreted as html mark-up tags
    

    launch.get('notify-send', '-t', int(timeout * 1000), title, text)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import DependencyManager

with DependencyManager(__name__) as manager:
    pynput = manager.import_module("pynput")

keyboard = pynput.keyboard.Controller()
DefinedKeys = pynput.keyboard.Key

text = keyboard.type

press = keyboard.press
rls = keyboard.release

names = {key.name: key for key in DefinedKeys}  # these names are defined
names["minus"]="-"
# in pynput
# alphanumeric keys are automatically supported via they standard names


# # support for numpad keys:
# numpad_to_normal = {'p_delete': 'Delete', 'p_home': 'Home', 'p_down':
# 'down', 'p_next':
# 'PageDown', 'p_add': 'plus',
#                     'p_right': 'right', 'p_page_up': 'PageUp',
# 'p_page_down': 'PageDown',
# 'p_multiply': 'multiply',
#                     'p_end': 'End', 'p_up': 'up', 'p_insert': 'insert',
# 'p_divide': 'slash',
#  'p_left': 'left',
#                     'p_subtract': 'minus'}
# keynames.update(numpad_to_normal)
# DOESNT WORK YET


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#



from .baseclasses import (InputEventHandler, KeyhookBase, ButtonhookBase,
    CursorhookBase, PressReleaseHook)
from ..event_classes import PressReleaseEvent

from .. import DependencyManager

with DependencyManager(__name__) as manager:
    device_control = manager.import_module("evdev_prepared.device_control")
    uinput = manager.import_module("evdev_prepared.uinput")


# the following import will not be performed if the DependencyManager is only
# checking the import capability. if evdev is not installed, an error will
# be raised when trying to import evdev_prepared already.
from evdev.ecodes import (EV_KEY, EV_ABS, EV_SYN, EV_MSC, KEY, BTN,
    EV_LED, EV_REL, ABS_MT_POSITION_X, ABS_MT_POSITION_Y, ABS_X, ABS_Y,
    REL_X, REL_Y, bytype)

class EvdevHandler(device_control.DeviceHandler, InputEventHandler):
    
   
    devupdater = device_control.DeviceUpdater()
    # this class will also use the CollactableInputDevice class and the
    # EvdevInputLooper class in the background
    uinput = uinput.global_uinput
    uinput.start()
    
    def __init__(self, hook_cls=None, **selection_kwargs):
        InputEventHandler.__init__(self,hook_cls)
        device_control.DeviceHandler.__init__(self,devupdater= self.devupdater,
                **selection_kwargs)
    
    def process_event(self, ty, co ,val, dev):
        if ty == EV_SYN: return
        press = bool(val != 0)
            # val==0 means that it was released, val==1 or
            # val==2 means it is pressed
        name = bytype[ty][co]
        if isinstance(name, (list,tuple)): name = name[0]
        self.queue_event(name.lower(), press=press)
    
    
    
    

class KeyHandler(EvdevHandler):
    
    reinject_implemented = True
    
    
    name_dict = {}
    for a, b in KEY.items():
        if isinstance(b, list): b = b[0]
        b = b[4:].lower()
        name_dict[a] = b
            
    def process_event(self,ty,co,val, dev):
        #_print(ty,co,val,dev)
        if ty == EV_KEY:
            press = bool(val != 0)
            if self.queue_event(co, press=press) is True:
                self.uinput.write(ty, co, val)
        elif ty not in (EV_SYN, EV_MSC, EV_LED):
            raise TypeError("Wrong event type: ", ty, co , val)
        
        

class ButtonHandler(EvdevHandler):
    
    capture_allowed = False
    
    name_dict = {}
    for a, b in BTN.items():
        if isinstance(b, list): b = b[0]
        b = b[4:].lower()
        name_dict[a] = b

    def process_event(self, ty, co, val, dev):
        #print(ty,co,val)
        if ty == EV_KEY:
            press = bool(val != 0)
            self.queue_event(co, press=press)
        elif ty not in (EV_SYN, EV_MSC, EV_ABS, EV_REL):
            raise ValueError("Wrong event type: ", ty, co ,val)


class CursorHandler(EvdevHandler):
    
    capture_allowed = False

    def __init__(self, *args,**kwargs):
        super().__init__(*args,**kwargs)
        self.saved_vals = {}
        
    def _intelligent_collect(self, co ,co_x, co_y, val, **kwargs):
        if co == co_x:
            try:
                pos_old = self.saved_vals.pop(co_y)
            except KeyError:
                self.saved_vals[co_x] = val
            else:
                self.queue_event(val, pos_old, **kwargs)
                return True
        elif co == co_y:
            try:
                pos_old = self.saved_vals.pop(co_x)
            except KeyError:
                self.saved_vals[co_y] = val
            else:
                self.queue_event(pos_old, val, **kwargs)
                return True
            
    def process_event(self, ty, co, val, dev):
        # print(ty,co,val)
        if ty == EV_KEY:
            pass
        elif ty == EV_ABS:
            self._intelligent_collect(co, ABS_X, ABS_Y, val,
                    screen_coordinates=False)
        elif ty == EV_REL:
            self._intelligent_collect(co,REL_X, REL_Y, val, relative=True,
                    screen_coordinates=False)
        elif ty not in (EV_SYN, EV_MSC):
            raise ValueError("Wrong event type: ", ty, co, val)


class EvdevhookMixin:
    
    def process_custom_kwargs(self,  **selection_kwargs):
        # selection_kwargs can be name, category, path, see evdev_prepared
        if not selection_kwargs: return
        # create a new collector or capturer instance with updated
        # selection_kwargs
        old = self.handler
        cls = EvdevHandler if not old else old.__class__
        new = cls(self.__class__, **selection_kwargs)
        if old != new: self.handler = new
        
            
    def matching_devs(self):
        return self.handler.matching_devs()
    
    def collected_devs(self):
        return self.handler.collected_devs
    
    def captured_devs(self):
        return self.handler.grabbed_devs



class Keyhook(EvdevhookMixin, KeyhookBase):
    handler = KeyHandler(category="keyboard")

    #name_translation_dict = {"braceleft": "bracketleft", "braceright":
     #   "bracketright"}
    

class Buttonhook(EvdevhookMixin, ButtonhookBase):
    handler = ButtonHandler(category="mouse")

class Cursorhook(EvdevhookMixin, CursorhookBase):
    handler = CursorHandler(category="mouse")


class Customhook(EvdevhookMixin, PressReleaseHook):
    
    EventClass = PressReleaseEvent
    
    # collector is set via process_custom_kwargs (see above)
    
    def process_custom_kwargs(self, **selection_kwargs):
        if not selection_kwargs:
            selection_kwargs = {"exclude_category": ("mouse", "keyboard")}
        super().process_custom_kwargs(**selection_kwargs)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import functools
from .. import Adaptor, AdaptionError, adaptionmethod, Layout, DependencyManager

_evdev_infos = {"pkg": "evdev_prepared",
        "instruction": "Before using, please follow steps at "
                        "https://github.com/dp0s/evdev_prepared#install",
        "platform": "linux"
         }

DependencyManager.save_dependency_info("evdev_prepared.device_control",
        **_evdev_infos)
DependencyManager.save_dependency_info("evdev_prepared.uinput",**_evdev_infos)
DependencyManager.save_dependency_info("pynput",pkg="pynput")


class hotkeys:
    @classmethod
    def add_pause_option(cls, default_pause_value, timeout=10):
        def wrapper_function(func):
            @functools.wraps(func)
            def function_with_hotkey_pause(*args, pause_hotkeys=default_pause_value, **kwargs):
                if pause_hotkeys is True:
                    with cls.paused(timeout): return func(*args, **kwargs)
                else:
                    return func(*args, **kwargs)
            #function_with_hotkey_pause.__signature__ = add_kw_only_arg(func,
             #       "pause_hotkeys", default=default_pause_value)
            return function_with_hotkey_pause
        return wrapper_function
    
    def paused(self,*args,**kwargs):
        raise Exception("paused called too early")

    def TriggerManager(self,*args,**kwargs):
        raise Exception("TriggerManager called too early")

# a container for the TriggerManager switch
# this is necessary to avoid circular imports between modules
# this must be defined before the imports,
# because several modules need to import this
# the content is set later after the TriggerManager is defined

from .keybutton_classes import *
from .keybutton_names import *

Adaptor.NamedButtonClass = NamedButton
Adaptor.NamedKeyClass = NamedKey


from .event_classes import *




from .hookpower import *

hook = HookAdaptor(_primary_name="hook")

from .sending import *

mouse = MouseAdaptor(_primary_name="mouse")
keyb = KeyboardAdaptor(_primary_name="keyb")

from .event_classes import *



from .sending.event_sender import *

from .waiter import *
from .trigman import *

hotkeys.TriggerManager = TriggerManager
hotkeys.paused = TriggerManager.paused
# usage: with hotkeys.paused(time)



del Adaptor, adaptionmethod, AdaptionError


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import time, os, sys, threading, traceback, warnings, functools
from Dhelpers.container import container
from Dhelpers.builtin_manipulation import execute_after_print,\
    execute_after_warnings, execute_after_error, always_print_traceback
#the last two print functions should be imported via the Dfuncs module
from Dhelpers.arghandling import PositiveInt
from .winapps import pythoneditor

from .events import KeyboardAdaptor, MouseAdaptor, KeyWaiter, HookAdaptor, CombinedSender
from . import dlg, ntfy, clip, Win, launch
from .windowpower import WindowObject

# =======================================================================
# SENDWAIT function
# =======================================================================

keyb = KeyboardAdaptor(group="Dfuncs",_primary_name="keyb")
mouse = MouseAdaptor(group="Dfuncs", _primary_name="mouse")
event_sender = CombinedSender(keyb,mouse)

class SendwaitError(Exception):
    pass
        

def nfsendwait(*args, **kwargs):
    """function shortcut to enable notifications"""
    return sendwait(*args, nfOn=True, notifytime=20, **kwargs)


def sendwait(mandatory, *optional, winwaittime=10, notifytime=5,
        continueOnError=False, nfOn=False, delay=None):
    """A multi-purpose function to quickly pass a chain of commands:
    The default symbol "→" is used to escape special command tags.
    Functions see code.
    """
    
    def report(a, b):
        if nfOn:
            ntfy(a, notifytime, b)
    
    args = (mandatory,) + optional
    
    for i in range(len(args)):
        ai = args[i]
        # notify(str(ai))
        error = None
        
        if isinstance(ai, (int, float)):
            report("sleeping", ai)
            time.sleep(ai)
        
        elif isinstance(ai, str):
            report("Sending keys:", ai)
            event_sender.send(ai)
        
        elif isinstance(ai, WindowObject):
            report("Waiting for Window to exist and activate:", ai)
            found_win = ai.wait_exist_activate(timeout=winwaittime)
            if found_win:
                report("Found active window:", ai)
            else:
                error = "Window did not exist after winwaittime"
        else:
            try:
                event_sender.send_event(ai)
            except TypeError:
                pass
            else:
                report(f"Sending event:", ai)
            try:
                pid = ai.pid
            except AttributeError:
                error = "Wrong argument type for startwait command."
            else:
                if isinstance(pid, PositiveInt):
                    report("Waiting for Window to exist and activate:",
                            "pid={pid}\nProcessObject='{ai}'".format_map(
                                    locals()))
                    found_win = Win(pid=pid).wait_exist(timeout=winwaittime)
                    if found_win:
                        report(f"Found active window: pid"
                            f"={pid}\nProcessObject='{ai}'")
                    else:
                        error = "Window did not exist after winwaittime"
                else:
                    error = "Wrong argument type for startwait command."
        
        # checking if there was an error:
        if error is not None:
            if continueOnError:
                # Change the behavior if startwait encounters an error.
                # Default= Notify and Exit
                return error
            else:
                raise SendwaitError(f"Argument {i} gave the following error: "
                    f"{error}")
        if delay: time.sleep(delay)
    # after looping over all arguments:
    return True



# ========================================================
### Catching messages to additionally show a notification
# ========================================================

def notify_about_error(hotkey=None, timeout=10):
    @execute_after_error
    def error_notification(error_type, error_message, error_traceback):
        source_msg = ""
        if hotkey: source_msg += f"Press {hotkey} to see."
        ntfy(f"{error_type.__name__}: {error_message}", timeout, source_msg)
        def analyze_error():
            options = traceback.format_tb(error_traceback)
            text = f"{error_type}\n{error_message}\n\nTraeback:\n"
            result = dlg.choose(options, options[-1], text=text, width=1000,
                    height=1000)
            if result:
                i = options.index(result)
                fs = traceback.extract_tb(error_traceback, i + 1)[i]
                pythoneditor.jump_to_line(fs[0], fs[1])
        container.set_temp_store_key("exec_func",analyze_error, timeout)



def notify_about_warnings(hotkey=None,timeout=10):
    @execute_after_warnings
    def post_notifcation(msg,file,line):
        source_msg = f"Printed on line {line} in: {file}."
        if hotkey: source_msg+= f" Press {hotkey} to see."
        ntfy(f"Warning: {msg}", timeout, source_msg)
        container.set_temp_store_key("exec_func", functools.partial(
                pythoneditor.jump_to_line,file, line), timeout)
        

def notify_about_print(hotkey=None, timeout=5):
    @execute_after_print
    def ntfy_after_print(*args, sep=" ", end="\n", notify=True, **ignore):
        if not notify: return
        file, line, func, text = traceback.extract_stack(limit=3)[-3]
        # there must be at least three frames: this one, _custom_print,
        # call of print func
        # we want to access the call of the print_func, so we use [-3]
        print_message = sep.join([str(arg) for arg in args])
        if end != "\n": print_message += end
        source_msg = f"Printed on line {line} in: {file}."
        if hotkey: source_msg += f" Press {hotkey} to see."
        ntfy(print_message, timeout, source_msg)
        container.set_temp_store_key("exec_func", functools.partial(
                pythoneditor.jump_to_line, file, line), timeout*2)



# ===============================================================

def display_sorted_dict(dic, name=""):
    f = ""
    for i in sorted(dic):
        #    notify(i)
        f += i + "\t"*(3 + (i[0].upper() == "L") - len(i)//4) + "-> " + dic[
            i] + "\n"
        # notify(i)
    launch.get("echo", f, "|", "zenity", "--text-info")

extdict = {"jpeg": "jpg", "tiff": "tif"}


class FilePathParts:
    def __init__(self, file_path):
        self.path = file_path.rstrip("/")
        self.dir, self.base = os.path.split(self.path)
        self.name, self.ext = os.path.splitext(self.base)
        if self.ext == "":
            if os.path.isdir(self.path):
                self.ext = "dir"
        a = self.ext.lower()
        self.ext_norm = extdict.get(a,
                a)  # replace extension by version in extdict if present


# =============================================================================
#     Displaying window information and putting it to clipboard
# =============================================================================

def display_win_info():
    ntfy("Click on a window", 3)
    
    x = Win(loc="SELECT").all_info()
    winprops = x[:3] + ((x[1], x[2]),) + x[3:]
    
    show = [str(winprops[0]) + " [ID]", str(winprops[1]) + " [TITLE]",
        str(winprops[2]) + " [CLASS]", str(winprops[3]),
        str(winprops[4]) + " [PID]",
        str(winprops[5]) + " [GEOMETRY] (x,y,width,height)"]
    
    ret = dlg.choose(show, default=3, title="Window information",
            text="Save to clipboard:", width=700)
    
    if ret is not None:
        for i in range(len(show)):
            # dlg.msg(str(ret)+"\n"+str(show[i]))
            if ret == show[i]:
                clip.fill(winprops[i], notify=True)
                break


# =============================================================================
#     Display key names
# =============================================================================
def display_key_names():
    y = KeyWaiter.get1key()
    x = tuple(y.named_instance.names)
    time.sleep(0.1)
    if dlg.quest("Dpowers key name(s):\n{x}\n\nSave first(=standard) name to "
                 "clipboard?".format_map(locals()),title="Get key names"):
        clip.fill(x[0], notify=True)
        
def monitor_input_events(timeout=10, implementation=None,**hookkwargs):
    if implementation:
        this_hook=HookAdaptor(implementation)
    else:
        this_hook=HookAdaptor(group="keywait")
    ntfy("Showing all input events for %s seconds." % timeout)
    with this_hook.keyboard_mouse(functools.partial(ntfy,timeout=1), **hookkwargs):
        time.sleep(timeout)
    ntfy("Done showing input events.")


saved_patterns = []
    
def save_key_replay():
    k = KeyWaiter(release=True, buttons = True, maxtime=600,endevents=("Esc",))
    ntfy("Saving pressed keys",10, "Abort by pressing Esc")
    saved_patterns.append(k)
    k.start()
    ntfy("Stop saving pressed keys")

def replay_pressed_keys(delay=100):
    k = saved_patterns[-1]
    ntfy("Press ShiftL to start")
    if KeyWaiter.wait_for_key("ShiftL"):
        ntfy("reinjecting")
        k.reinject(delay=delay)
        ntfy("reinjecting done")


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import DependencyManager

with DependencyManager(__name__) as tester:
    mp3 = tester.import_module("eyed3.mp3", pkg="eyed3")

import eyed3 #this is already imported and will just place the name here again

obj_class = eyed3.id3.tag.Tag
eyed3.log.setLevel("ERROR")

def load(file, **kwargs):
    file = eyed3.load(file)
    return file.tag
    

def save(backend_obj, destination):
    backend_obj.save(destination)

def close(backend_obj):
    pass

def genre(backend_obj, val):
    if val is None:
        g = backend_obj.genre
        if g is None: return ""
        return g.name
    backend_obj.genre = val
    



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from .. import Win, launch, DependencyManager
import os


with DependencyManager(__name__) as manager:
    manager.test_shellcommand("zenity", pkg="zenity")

def run_zenity(arg1, *args, allow_retCodes=(0, 1), check_err=False):
    cmd = ("zenity", "--" + arg1, *args)
    # print(cmd)
    stderr = True if check_err else None
    process = launch(cmd, stdout=True, stderr=stderr)
    Win(pid=process.pid).wait_exist_activate()
    out, err = process.communicate()
    retCode = process.returncode
    if retCode != 0 and retCode not in allow_retCodes:
        raise launch.CalledProcessError(retCode, process.args)
    if check_err and err:
        if err != "Gtk-Message: GtkDialog mapped without a transient parent. " \
                  "This is discouraged.\n":
            raise launch.SubprocessError(
                    "Subprocess {process.args} returned error:\n{"
                    "err}".format_map(locals()))
    if out.endswith("\n"): out = out[:-1]
    return out, err, retCode


def kwargs_helper(**kwargs):
    kwargs["no_markup"] = kwargs.get("no_markup", True)
    # we want to use the no_markup_option by default to make sure that the
    # zenity ability for pango markup is disabled by default.
    for param, value in kwargs.items():
        if "_" in param:
            param = param.replace("_", "-", 1)
        if value is True:
            # print(param,value)
            yield "--" + param
        elif value not in (None, False):
            yield "--%s=%s"%(param, str(value))



def date(selected, **kwargs):
    if selected:
        kwargs["day"], kwargs["month"], kwargs["year"] = selected
    out, err, retCode = run_zenity("calendar",
            *kwargs_helper(date_format="%d/%m/%Y", **kwargs))
    if retCode == 0: return tuple(int(x) for x in out.split("/"))


def path(selected, multi, dir, **kwargs):
    if selected:
        if not os.path.lexists(selected):
            raise ValueError("File %s does not exist!"%selected)
        kwargs['filename'] = selected
    out, err, retCode = run_zenity('file-selection',
            *kwargs_helper(multiple=multi, separator="|", directory=dir,
                    **kwargs))
    if retCode == 0: return out.split("|") if multi else out


def savepath(selected, overwrite, **kwargs):
    out, err, retCode = run_zenity('file-selection',
            *kwargs_helper(filename=selected, save=True,
                    confirm_overwrite=not overwrite, **kwargs))
    if retCode == 0: return out


def msg(text, title, error, warning, **kwargs):
    if error:
        cmd = "error"
    elif warning:
        cmd = "warning"
    else:
        cmd = "info"
    out, err, retCode = run_zenity(cmd,
            *kwargs_helper(text=text, title=title, **kwargs))
    if retCode == 0: return out


def quest(text, title, **kwargs):
    out, err, retCode = run_zenity("question",
            *kwargs_helper(text=text, title=title, **kwargs))
    return retCode == 0


def inp(text, title, **kwargs):
    out, err, retCode = run_zenity("entry",
            *kwargs_helper(text=text, title=title, **kwargs))
    if retCode == 0: return out


def table(column_names, data, autoheight=True, text=None, title=None,
        select_col=None, multi=False, **kwargs):
    if column_names.count("") == len(column_names):
        # this is true if column_names only contains empty strings
        kwargs["hide_header"] = True
    if autoheight:
        pixel = 160
        pixel += (len(data) - 1)//len(column_names)*28
        if text:
            pixel += text.count("\n")*20
        kwargs["height"] = pixel
    if select_col: kwargs["print_column"] = True
    
    args = []
    for column in column_names: args.append('--column=%s'%column)
    
    args += list(kwargs_helper(text=text, title=title, separator="|",
            checklist=multi, **kwargs))
    
    for datum in data:
        args.append(str(datum))
    
    out, err, retCode = run_zenity("list", *args)
    if retCode == 0: return out.split("|") if multi else out



def choose(options, defaults, text, title, multi, **kwargs):
    data = []
    for i in range(len(options)):
        x = options[i]
        if not isinstance(x, str):
            raise TypeError("Options must all be of type str.")
        data.append((i in defaults) or (x in defaults))
        # This determines the default values marked.
        data.append(x)
    if not multi: kwargs["radiolist"] = True
    return table(column_names=("", ""), data=data, autoheight=True, text=text,
            title=title, multi=multi, **kwargs)



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from .. import Adaptor, adaptionmethod
from .windowobjects import FoundWindows, WindowSearch, WindowObject
from collections import defaultdict
import functools
from contextlib import contextmanager

class WindowAdaptor(Adaptor):
    
    
    @functools.wraps(Adaptor.__init__)
    def __init__(self, *args,**kwargs):
        super().__init__(*args,**kwargs)
        self.cached_properties = defaultdict(dict)
        self.use_cache = False
        
    @functools.wraps(Adaptor.adapt)
    def adapt(self, *args,**kwargs):
        self.cached_properties.clear()
        return super().adapt(*args,**kwargs)


    @contextmanager
    def cached(self):
        self.cached_properties.clear()
        self.use_cache  =True
        try:
            yield
        finally:
            self.use_cache = False
            self.cached_properties.clear()


    @adaptionmethod
    def screen_res(self):
        return self.screen_res.target()
    
    @adaptionmethod("ID_from_location")
    def _ID_from_location(self, param):
        try:
            i = self._ID_from_location.target_with_args()
        except NotImplementedError as e:
            raise NotImplementedError(f"The adaption module "
              f" {self._ID_from_location.module} does not support the following"
              f" location argument for window objects: {param}") from e
        return i
    
    
    @staticmethod
    def _error(prop_name, func):
        return NotImplementedError(
                f"\nTried to use the following property name for window "
                f"argument: {prop_name}\nHowever, the following adaption "
                f"module does not support this window property name:\n"
                f"{func.__module__}")
    
    @adaptionmethod(require=True)
    def IDs_from_property(self, prop_name, prop_val, visible=None):
        try:
            i_list = self.IDs_from_property.target_with_args()
        except NotImplementedError as e:
            raise self._error(prop_name,self.IDs_from_property) from e
        if not i_list: i_list = set()
        if self.use_cache:
            for id in i_list: self.cached_properties[id][prop_name]=prop_val
        return i_list
    
    @adaptionmethod(require=True)
    def property_from_ID(self, ID, prop_name, query_cache=None):
        if query_cache is None and self.use_cache is True: query_cache = True
        if query_cache:
            try:
                return self.cached_properties[ID][prop_name]
            except KeyError:
                pass
        try:
            val = self.property_from_ID.target(ID, prop_name)
        except NotImplementedError as e:
            raise self._error(prop_name,self.property_from_ID) from e
        if not val: val = None
        if self.use_cache: self.cached_properties[ID][prop_name]=val
        return val
    
    def id_exists(self, ID):
        return bool(self.property_from_ID(ID, "title", query_cache=False))
    
    @adaptionmethod
    def activate(self, ID):
        return self.activate.target_with_args()
    
    @adaptionmethod
    def map(self, ID):
        return self.map.target_with_args()

    @adaptionmethod
    def unmap(self, ID):
        return self.unmap.target_with_args()
    
    @adaptionmethod
    def set_prop(self, ID, action: str, prop: str, prop2: str = False):
        return self.set_prop.target_with_args()
    
    @adaptionmethod
    def move(self, ID, x=-1, y=-1, width=-1, height=-1):
        return self.move.target_with_args()
    
    @adaptionmethod
    def close(self, ID):
        return self.close.target_with_args()
    
    @adaptionmethod
    def kill(self, ID):
        return self.kill.target_with_args()
    
    @adaptionmethod
    def minimize(self, ID):
        return self.minimize.target_with_args()


    
class WindowHandler(FoundWindows, WindowAdaptor.AdaptiveClass):
    pass
    
    
    


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from ... import adaptionmethod, hotkeys, NamedKey, Layout
from ..event_sender import AdaptivePRSenderShifted


class KeyboardAdaptor(AdaptivePRSenderShifted):
    """Send key events to the system via the chosen backend.
    
    See `Dpowers.events.keybutton_names.py
    <https://github.com/dp0s/Dpowers/tree/master/Dlib/Dpowers/events
    /keybutton_names.py>`_ for a list of allowed key names and key groups.
    Most of the defined keys can be called by more than one name equivalently.
    Simply pass one of the key's names as a string to the methods defined below.
    
    For some backends (such as *evdev*) it might be neccessary to manually
    specify the layout of the keyboard currently used via
    the :func:`set_layout` method.
    """
    
    
    #Following adaptionmethods are mandatory in adapt_*.py:
    # press
    # rls
    
    @adaptionmethod("text")
    def _text_action(self, character, **kwargs):
        self._text_action.target(character, **kwargs)
    
    default_delay = 0
    """
    .. exec::

        event_sender.default_delay_doc("keys")
    """

    default_duration = 1
    """
    .. exec::

        event_sender.default_duration_doc("key")
    """

    @property
    def NamedClass(self):
        return self.NamedKeyClass
    @NamedClass.setter
    def NamedClass(self, val):
        if not issubclass(val, NamedKey): raise TypeError
        self.NamedKeyClass = val
        self.create_effective_dict()
            
        

    @hotkeys.add_pause_option(True)
    def send(self, string, auto_rls=True, delay=None):
        """
        Convenience method to send several kind of key patterns in one
        command. It integrates the functionality of :func:`tap`, :func:`comb`
        and :func:`text`.

        :param str string: Content to be sent, format see below.
        :param bool auto_rls: If ``True``, send press and release events for
            all keys specified inside ``< >``. If ``False``, normal key names
            are
            interpreted as press events and you must attach ``_rls`` to a key
            name to send the corresponding release event.
        :param int delay: Time (in milliseconds) to wait between tapping
            keys. Defaults to :attr:`default_delay`.

        The **string** parameter will be parsed from left to right according to
        the following rules:

        - There are two special characters: ``<`` and ``>``.
          All content between them will be interpreted as key
          names (as if called by the :func:`tap` method)::

            keyb.send("<return>") # is equivalent to
            keyb.tap("return")

        - Everything outside of a ``< >`` block will be sent literally using
          the :func:`text` method::

            keyb.send("normal text input..") # is equivalent to
            keyb.text("normal text input..")

        - Use ``<key1+key2>`` to send key combinations::

            keyb.send("<ctrl+s>") # is equivalent to
            keyb.comb("ctrl","s") # is equivalent to
            keyb.send("<ctrl s s_rls ctrl_rls>", auto_rls=False)
        - Chain keys and texts as in the following examples::

            keyb.send("<home>hello world<return end>") # equivalent to
            keyb.tap("home","h","e","l","l","o"," ","w","o","r","l","d",
            "return", "end")

            keyb.send("<key1 key2 key3>") # equivalent to
            keyb.send("<key1><key2><key3>")

            keyb.send("<shift+home backspace>this line was deleted")
            # is equivalent to
            keyb.comb("shift","home")
            keyb.tap("backspace")
            keyb.text("this line was deleted")

        - If not paired, ``<`` and ``>`` are interpreted literally.
        - Use ``<><something>`` to send literal *<something>* as text output.

        """
        super().send(string, auto_rls=auto_rls, delay=delay)


    
    @hotkeys.add_pause_option(True)
    def tap(self, key, *keys, delay=None, duration=None, repeat=1):
        """Simulate a key tap (i.e. send a press event followed by a release
        event of the same key), or multiple key taps in a row.

        :param str key: Name of the (first) key to be sent.
        :param str keys: Further names of keys to be sent in sequence.
        :param int delay: Time (in milliseconds) to wait after tapping a key.
            Defaults to :attr:`default_delay`.
        :param int duration: Time (in milliseconds) to wait between each press
            and release event. Defaults to :attr:`default_duration`.
        :param repeat: Number of times this sequence should be repeated.

        Examples::
        
            keyb.tap("a")
            keyb.tap(1)
            keyb.tap("esc")
            keyb.tap("f9")
        """
        return super().tap(key, *keys, delay=delay, duration=duration,
                repeat=repeat)

    @hotkeys.add_pause_option(True)
    def press(self, key, *keys, delay=None, translate_names = True):
        """Send key press event(s) to the system. You need to manually send
        the corresponding release event(s) afterwards.

        :parameters: See :func:`tap`.
        """
        return super().press(key, *keys, delay=delay)

    @hotkeys.add_pause_option(True)
    def rls(self, key, *keys, delay=None):
        """Send key release event(s) to the system.

        :parameters: See :func:`tap`.
        """
        return super().rls(key, *keys, delay=delay)
    


    @hotkeys.add_pause_option(True)
    def comb(self, key1, key2, *keys, delay=None, duration=None):
        """Simulates a key combination, such as *Control+S* or *Shift+Alt+F5*.
        Sends all the press events first and then the release events in
        reverse order.
        
        :param str key1: The is the name of the first key to be pressed (and
            hence the last to be released.)
        :param str key2: The name of the second key in the combination.
        :param str keys: Further key names.
        :param int delay: Time (in milliseconds) to wait between
            pressing/releasing the sequence of keys. Defaults to
            :attr:`default_delay`.
        :param int duration: Time (in milliseconds) to wait between sending
            the press and release event of the final key. Defaults to
            :attr:`default_duration`.
        
        Examples::
        
            keyb.comb("ctrl","s")
            keyb.comb("shift","alt","f5")
        """
        return super().comb(key1, key2, *keys, delay=delay, duration=duration)

    custom_text_method = True
    """Default value for parameter **custom** of method :func:`text`."""
    


    @hotkeys.add_pause_option(True)
    def text(self, text, delay=None, custom=None, **kwargs):
        """Send a plain text string to the system as if the user would type it.
        
        :param str text: Any text you want to send. Depending on the backend,
             not all symbols might be supported.
        :param int delay: Time (in milliseconds) to wait between sending
            single characters. Defaults to :attr:`default_delay`.
        :param bool custom: If set to ``True``, prefer the backend's special text
            method if implemented. Otherwise use :func:`tap` for sending
            single characters via the backend's press/rls methods.
        :param kwargs: Custom keyword arguments to pass to the backend's
            text method.
        """
        return super().text(text,delay=delay, custom=custom, **kwargs)

    _pa = AdaptivePRSenderShifted._press_action
    backend_layout = _pa.adaptive_property("backend_layout", ty=(Layout, str))
    layout = _pa.adaptive_property("layout", ty=(Layout, str))
    """Return the currently used keyboard layout for this instance. Use
    :func:`set_layout` to change it manually."""
    
    def set_layout(self, name, make_default=False, backend_layout=None,
            use_shifted=None):
        """Manually set the effective keyboard layout and
        translate the key codes from the backend accordingly. This is e.g.
        necessary for the *evdev* backend to make sure that the correct keys
        are sent to the system.

        :param str name: Name of the keyboard layout you use. This is usually a
            two character abbreviation, such as *'de'* or *'us'*. A list of
            available layouts can be found in the folder
            `Dhelpers.KeyboardLayouts.layouts_imported_from_xkb
            <https://github.com/dp0s/Dpowers/tree/master/Dlib/Dhelpers
            /KeyboardLayouts/layouts_imported_from_xkb>`_.
        :param bool make_default: If set to ``True``, all other instances of
            :class:`KeyboardAdaptor` with the same backend as this instance will
            use this layout by default. Raises an exception if the backend
            for this instance has not been chosen yet.
        :param str backend_layout: Manually set the assumed backend
            layout. Normally this is set by the backend, so don't use this
            parameter unless you know what you are doing.
        :param bool use_shifted:  Whether to use manualy key shifting when
            sending keys. Normally this is set by the backend, so don't use this
            parameter unless you know what you are doing.
        
        Example for a German keyboard:
        
        .. activecode::
            
            > from Dpowers import keyb
            > keyb.adapt("evdev")
            "<module 'Dpowers.events.sending.keybpower.adapt_evdev'>"
            > keyb.layout # returns default layout for evdev backend
            "us"
            > keyb.tap("[") # sents the wrong key on a German keyboard:
            > ü
            > # change to German layout & set as default for evdev backend:
            > keyb.set_layout("de", make_default=True)
            > keyb.tap("[") # now the correct key is sent:
            > [
        
        """
        self.translation_dict = None
        self.backend_layout = backend_layout
        self.use_shifted = use_shifted
        self.layout = name
        self.create_effective_dict(make_default=make_default)

    def set_translation_dict(self, dic, make_default=False):
        # idea: combine translation dict and layout
        self.layout = False
        self.backend_layout = False
        super().set_translation_dict(dic, make_default)


    def _create_effective_dict(self, ts, enforced_inst=None):
        if enforced_inst:
            self.layout = enforced_inst.layout
            self.backend_layout = enforced_inst.backend_layout
        if self.layout:
            if self.backend_layout is None:
                raise ValueError(
                        "Could not determine backend layout for backend"
                        f" {ts}. Please set it manually.")
            L = Layout(self.layout)
            self.translation_dict = L.send_map(to=self.backend_layout,
                    shifted=self.use_shifted)
            enforced_inst = None  # this way, we avoid enforcing it twice
        super()._create_effective_dict(ts, enforced_inst)
    
    
    @property
    def key(self):
        """This object allows accessing Dpowers' internal key objects and key
        groups as defined in the module `Dpowers.events.keybutton_names.py
        <https://github.com/dp0s/Dpowers/tree/master/Dlib/Dpowers/events
        /keybutton_names.py>`_. Find a key object by calling it's name:

        .. activecode::
            
            > keyb.key("a")
            repr(keyb.key("a"))
            > keyb.key("ctrl")
            repr(keyb.key("ctrl"))
            > keyb.key(1)
            repr(keyb.key(1))
            > # check if two names belong to the same key object:
            > keyb.key("ctrl") is keyb.key("leftcontrol")
            keyb.key("ctrl") is keyb.key("leftcontrol")

        You can find key groups via the **.group** attribute and check if a
        key belongs to it:

        .. activecode::
        
            > keyb.key.group.arrow_keys
            keyb.key.group.arrow_keys
            > "up" in keyb.key.group.arrow_keys
            True
            > 0 in keyb.key.group.arrow_keys
            False
            > 0 in keyb.key.group.digits
            True
            
        """
        return self.NamedKeyClass.NameContainer